def pH_adjustment(
    Path,
    DesiredVariables,
    Dates, 
    Cant, 
    Cant2002,
    PerKgSwTF, 
    Cant_adjusted={},
    Est_pre={}, 
    PredictorMeasurements={}, 
    OutputCoordinates={}, 
    C={},  
    Uncertainties_pre={}, 
    DUncertainties_pre={},
    **kwargs
):

    """
    Adjusting pH for anthropogenic carbon in the second of three Cant adjustment
        scripts

    Inputs:
        Path: User-defined computer path
        DesiredVariables: List of desired variables to estimate
        Dates: List of dates for measurements
        Cant: Numpy array of anthropogenic carbon estimates for each estimate
        Cant2002: Numpy array of anthropogenic carbon estimates for each
            estimate for the year 2002
        PerKgSwTF: Boolean for whether measurements were in molar or molal units
        Cant_adjusted: Dictionary of estimates adjusted for Cant
        Est_pre: Dictionary of estimates for each combination
        PredictorMeasurements: Dictionary of measurements for estimates
        OutputCoordinates: Dictionary of geographic coordinates for estimates
        C: Dictionary of adjusted coordinates for estimates
        Uncertainties_pre: Dictionary of measurement uncertainties for estimates
        DUncertainties_pre: Dictionary of default measurement uncertainties for
            estimates
        **kwargs: Please see README

    Output:
        Cant_adjusted: Dictionary of adjusted values for Cant for each combination
    """

    import numpy as np
    import seawater as sw
    import PyCO2SYS as pyco2
    from PyESPER.inputdata_organize import inputdata_organize
    from PyESPER.temperature_define import temperature_define
    from PyESPER.iterations import iterations
    from PyESPER.fetch_data import fetch_data
    from PyESPER.input_AAinds import input_AAinds
    from PyESPER.coefs_AAinds import coefs_AAinds
    from PyESPER.interpolate import interpolate
    from PyESPER.organize_data import organize_data
        
    # Finding the estimates within the dictionary
    combos2 = list(Est_pre.keys())
    values2 = list(Est_pre.values())
        
    # Setting the default warning message boolean and Verbose
    YouHaveBeenWarnedCanth = False
    VerboseTF = kwargs.get("VerboseTF", True)

    # Proceed only if dates are provided and DIC or pH is requested
    if "EstDates" in kwargs and ("DIC" in DesiredVariables or "pH" in DesiredVariables):
        if "pH" in DesiredVariables:
            warning = []
            for combo, values in zip(combos2, values2):
                if combo.startswith("pH"):
                    # Rerunning LIRs for when pH is requested to obtain TA estimates
                    # based only on salinity input
                    salinity = PredictorMeasurements["salinity"]
                    PM_pH = {"salinity": salinity}
                    eq = [16]
                    InputAll2 = inputdata_organize(
                        Dates,
                        C,
                        PM_pH,
                        Uncertainties_pre
                    )
                    PredictorMeasurements2, InputAll2 = temperature_define(
                        ["TA"],   
                        PM_pH,
                        InputAll2,
                        **kwargs
                    )
                    code, unc_combo_dict, dunc_combo_dict = iterations(
                        ["TA"],
                        eq,
                        PerKgSwTF,
                        C,
                        PredictorMeasurements2,
                        InputAll2,
                        Uncertainties_pre,
                        DUncertainties_pre
                    )
                    LIR_data = fetch_data(["TA"], Path)
                    AAdata, Elsedata = input_AAinds(C, code)
                    Gdf, CsDesired = coefs_AAinds(eq, LIR_data)
                    aaLCs, aaInterpolants_pre, elLCs, elInterpolants_pre = interpolate(
                        Gdf,  
                        AAdata,
                        Elsedata
                    )
                    alkest, _ = organize_data(
                        aaLCs,
                        elLCs,    
                        aaInterpolants_pre,
                        elInterpolants_pre,
                        Gdf,
                        AAdata,
                        Elsedata
                    )
                    EstAlk = np.array(alkest["TA16"])
                    EstAlk = np.transpose(EstAlk)
                    EstSi = EstP = [0] * len(EstAlk)
                    # Calculating pressure using the sw package
                    Pressure = sw.pres(OutputCoordinates["depth"], OutputCoordinates["latitude"])
                    Est = np.array(values)
                    Est = np.transpose(Est)
                    Est = Est[0]
                    temperature = np.array(PredictorMeasurements["temperature"])
                    # Now using CO2SYS to calculate DIC based on the above LIR results for TA
                    # CO2SYS calculations
                    kwargCO2 = {
                        "par1":EstAlk,
                        "par2":Est,
                        "par1_type":1,
                        "par2_type":3,
                        "salinity":salinity,  
                        "temperature":temperature,
                        "temperature_out":temperature,
                        "pressure":Pressure,
                        "pressure_out":Pressure,
                        "total_silicate":EstSi,
                        "total_phosphate":EstP,
                        "opt_total_borate":2}
                    Out = pyco2.sys(**kwargCO2)
                    DICadj = Out["dic"] + Cant - Cant2002
                    
                    # Using this DIC to calculate pH from CO2SYS
                    kwargCO2_2 = {
                        "par1":EstAlk,
                        "par2":DICadj,
                        "par1_type":1,
                        "par2_type":2,
                        "salinity":salinity,
                        "temperature":temperature,
                        "temperature_out":temperature,
                        "pressure":Pressure,
                       "pressure_out":Pressure,
                        "total_silicate":EstSi,
                        "total_phosphate":EstP,
                        "opt_total_borate":2}
                    OutAdj = pyco2.sys(**kwargCO2_2)
                    pHadj = OutAdj["pH"]
                        
                    # Check for convergence warnings
                    if np.isnan(pHadj).any():   
                        warning_message = (
                            "Warning: CO2SYS took >20 iterations to converge. The corresponding estimate(s) will be NaN. "
                            "This typically happens when ESPER_LIR is poorly suited for estimating water with the given properties "
                            "(e.g., very high or low salinity or estimates in marginal seas)."
                        )
                        warning.append(warning_message)
                    # Append to adjusted Cant dictionary
                    Cant_adjusted[combo] = pHadj.tolist()
                        
                # Print warnings if any
                if warning:
                    print(warning[0]) 
                        
    elif "EstDates" not in kwargs and ("DIC" or "pH" in DesiredVariables) and not VerboseTF and YouHaveBeenWarnedCanth:
        print("Warning: DIC or pH is a requested output but the user did not provide dates for the desired esimtates. The estimates "
            "will be specific to 2002.0 unless the optional EstDates input is provided (recommended).")

        YouHaveBeenWarnedCanth = True
                        
    return Cant_adjusted



def pH_DIC_nn_adjustment(Path, DesiredVariables, Estimates, YouHaveBeenWarnedCanth, OutputCoordinates={}, PredictorMeasurements={}, **kwargs):

    """
    Calculating the anthropogenic carbon component for DIC and pH, if needed

    Inputs:
        Path: Path for user computer
        DesiredVariables: List of variables for which estimates are requested
        Estimates: Dictionary of estimates for each equation-variable combination
        YouHaveBeenWarnedCanth: Boolean for avoiding repeat warnings
        OutputCoordinates: Dictionary of coordinates for requested estimates
        PredictorMeasurements: Dictionary of measurements for calculations
        **kwargs: Please see README for full information

    Outputs:
        Cant_adjusted: Dictionary of estimates adjusted for anthropogenic carbon
            for each equation-variable combination
    """

    from PyESPER.simplecantestimatelr import simplecantestimatelr
    import numpy as np
    from PyESPER.defaults import defaults
    from PyESPER.lir_uncertainties import measurement_uncertainty_defaults
    from PyESPER.inputdata_organize import inputdata_organize
    from PyESPER.temperature_define import temperature_define
    from PyESPER.iterations import iterations
    from PyESPER.define_polygons import define_polygons
    from PyESPER.run_nets import run_nets
    from PyESPER.process_netresults import process_netresults
    import seawater as sw
    import PyCO2SYS as pyco2

    # Interpreting kwargs and setting defaults if needed
    VerboseTF = kwargs.get("VerboseTF", True)
    n = max(len(v) for v in OutputCoordinates.values())
    if "EstDates" in kwargs:
        d = np.array(kwargs["EstDates"])
        EstDates = (
            [item for sublist in [kwargs["EstDates"]] * (n + 1) for item in sublist]
            if len(d) != n else list(d)
        )
    else:
        EstDates = [2002.0] * n

    # Predefining empty dictionary for results and parsing out estimates
    Cant_adjusted = {}
    combos2 = list(Estimates.keys())
    values2 = list(Estimates.values())

    # Only if dates are available and if needed, continue
    if "EstDates" in kwargs and ("DIC" in DesiredVariables or "pH" in DesiredVariables):
        if not YouHaveBeenWarnedCanth:   
            if VerboseTF:
                print("Estimating anthropogenic carbon for PyESPER_NN.")
            longitude = np.mod(OutputCoordinates["longitude"], 360)
            latitude = np.array(OutputCoordinates["latitude"])
            depth = np.array(OutputCoordinates["depth"])
            Cant, Cant2002 = simplecantestimatelr(EstDates, longitude, latitude, depth)
            YouHaveBeenWarnedCanth = True
    
        for combo, a in zip(combos2, values2):
            a = np.array(a, dtype=float)
            if combo.startswith("DIC"):
                a = np.where(np.isnan(a), "nan", a + Cant - Cant2002)
                Cant_adjusted[combo] = a.tolist()
            else:
                Cant_adjusted[combo] = a.tolist()
        
        # Extra perturbation for pH
        if "pH" in DesiredVariables:   
            warning = []
            for combo, values in zip (combos2, values2):
                if combo.startswith("pH"):
                    salinity = PredictorMeasurements["salinity"]
                    PM_pH = {"salinity": salinity}
                    eq = [16]
                    _, n, _, EstDates, C, PerKgSwTF, MeasUncerts = defaults(
                        ["TA"],
                        PM_pH,
                        OutputCoordinates,
                        **kwargs
                    )
                    U_pre, DU_pre = measurement_uncertainty_defaults(
                        n,
                        PM_pH,
                        MeasUncerts
                    )
                    InputAll = inputdata_organize(
                        EstDates,
                        C,
                        PM_pH,
                        U_pre
                    )
                    PM_pH, InputAll = temperature_define(
                        ["TA"],
                        PM_pH,
                        InputAll,
                        **kwargs
                    )
                    code, unc_combo_dict, dunc_combo_dict = iterations(
                        ["TA"],
                        eq,
                        PerKgSwTF,
                        C,
                        PM_pH,
                        InputAll,
                        U_pre,
                        DU_pre 
                    )
                    df = define_polygons(C)
                    EstAtl, EstOther = run_nets(
                        ["TA"],
                        eq,
                        code
                    )
                    alkest = process_netresults(
                        eq,
                        code,
                        df,
                        EstAtl,
                        EstOther
                     )
                    EstAlk = np.array(alkest["TA16"])
                    EstSi = EstP = [0] * len(EstAlk)
                    Pressure = sw.pres(OutputCoordinates["depth"], OutputCoordinates["latitude"])
                    Est = np.array(values)
                        
                    # CO2SYS calculations
                    Out = pyco2.sys(
                        par1=EstAlk,
                        par2=Est,
                        par1_type=1,
                        par2_type=3,
                        salinity=salinity,
                        temperature=PredictorMeasurements["temperature"],
                        temperature_out=PredictorMeasurements["temperature"],
                        pressure=Pressure,
                        pressure_out=Pressure,
                        total_silicate=EstSi,
                        total_phosphate=EstP,
                        opt_total_borate= 2
                    )
                    DICadj = Out["dic"] + Cant - Cant2002
                    OutAdj = pyco2.sys(
                        par1=EstAlk,
                        par2=DICadj,
                        par1_type=1,
                        par2_type=2,
                        salinity=salinity,
                        temperature=PredictorMeasurements["temperature"],
                        temperature_out=PredictorMeasurements["temperature"],
                        pressure=Pressure,
                        pressure_out=Pressure,
                        total_silicate=EstSi,
                        total_phosphate=EstP,
                        opt_total_borate=2
                    )   
                    pHadj = OutAdj["pH"] 
                    
                    if any(np.isnan(pHadj)):
                        warning_message = (
                            "Warning: CO2SS took >20 iterations to converge. The correcponding estimate(s) will be NaN. "
                            "This typically happens when ESPER_NN is pooorly suited for estimating water with the given properties "
                            "(e.g., very high or low salinity or estimates in marginal seas)."
                        )
                        warning.append(warning_message)
                    else:
                        pHadj = np.array(values)
                        
                    Cant_adjusted[combo] = pHadj.tolist()
                        
                # Print warnings if any
                if warning:
                    print(warning[0])  
                        
    elif "EstDates" not in kwargs and ("DIC" or "pH" in DesiredVariables) and VerboseTF and not YouHaveBeenWarnedCanth:
        print("Warning: DIC or pH is a requested output but the user did not provide dates for the desired estimates. The estimates will be specific to 2002.0 unless the optional EstDates input is provided (recommended).")
        YouHaveBeenWarnedCanth = True
                        
    if kwargs.get("pHCalcTF") and "pH" in DesiredVariables:
        if VerboseTF:
            print("Recalculating the pH to be appropriate for pH values calculated from TA and DIC.")
        for combo, pH_values in zip(combos2, values2):
            if combo.startswith("pH"):
                pH_adjcalc_Est = [(pH + 0.3168) / 1.0404 for pH in pH_values]
                Cant_adjusted[combo] = pH_adjcalc_Est
                        
    return Cant_adjusted




def organize_nn_output(Path, DesiredVariables, OutputCoordinates={}, PredictorMeasurements={},  **kwargs):

    """
    Iteratively calculate uncertainties

    Inputs:
        Path: User-defined computer path
        Desired Variables: List of variables requested
        OutputCoordinates: Dictionary of geographic coordinates
        PredictorMeasurements: Dictionary of predictor measurements
        **kwargs: Please see README

    Outputs:
        Uncertainties: Dictionary of uncertainty estimates for each
            equation-variable combination
    """

    import numpy as np
    from PyESPER.defaults import defaults
    from PyESPER.lir_uncertainties import measurement_uncertainty_defaults
    from PyESPER.inputdata_organize import inputdata_organize
    from PyESPER.temperature_define import temperature_define
    from PyESPER.iterations import iterations
    from PyESPER.define_polygons import define_polygons
    from PyESPER.run_nets import run_nets
    from PyESPER.process_netresults import process_netresults
    from PyESPER.emlr_nn import emlr_nn

    # Predefine output lists
    PD_final, DPD_final, Unc_final, DUnc_final = [], [], [], []
    emlr = []

    # Rerun nets for each variable
    for d, var in enumerate(DesiredVariables):
        Pertdv, DPertdv, Unc, DUnc = [], [], [], []
        var = [var] # Wrap single variable in a list
        keys = ["sal_u", "temp_u", "phosphate_u", "nitrate_u", "silicate_u", "oxygen_u"]

        Equations, n, VerboseTF, EstDates, C, PerKgSwTF, MeasUncerts = defaults(
            var,
            PredictorMeasurements,
            OutputCoordinates,
            **kwargs
        )
        Uncertainties_pre, DUncertainties_pre = measurement_uncertainty_defaults(
            n,
            PredictorMeasurements,
            MeasUncerts
        )
    
        InputAll = inputdata_organize( 
            EstDates,
            C,
            PredictorMeasurements,
            Uncertainties_pre
        )
    
        PredictorMeasurements, InputAll = temperature_define(
            var,
            PredictorMeasurements,
            InputAll,
            **kwargs
        )
            
        code, unc_combo_dict, dunc_combo_dict = iterations(
            var,
            Equations,
            PerKgSwTF,
            C,
            PredictorMeasurements,
            InputAll,
            Uncertainties_pre,
            DUncertainties_pre
        )

        df = define_polygons(C)
            
        EstAtl, EstOther = run_nets(var, Equations, code)
            
        Estimate = process_netresults(
            Equations,
            code,
            df,
            EstAtl,
            EstOther
        )
            
        EMLR = emlr_nn(
            Path,
            var,
            Equations,
            OutputCoordinates,
            PredictorMeasurements
        )
      
        emlr.append(EMLR)
        names = list(PredictorMeasurements.keys())
        PMs = list(PredictorMeasurements.values())

        # Replace "nan" with 0 in PMs using list comprehensions
        PMs_nonan = [[0 if val == "nan" else val for val in pm] for pm in PMs]
        
        # Transpose PMs_nonan
        PMs = np.transpose(PMs_nonan)
            
        PMs3, DMs3 = {}, {}
            
        for pred in range(len(PredictorMeasurements)):
            num_coords = len(OutputCoordinates["longitude"])
            num_preds = len(PredictorMeasurements)
            
            # Initialize perturbation arrays
            Pert = np.zeros((num_coords, num_preds))
            DefaultPert = np.zeros((num_coords, num_preds))
            
            # Populate perturbation arrays
            Pert[:, pred] = Uncertainties_pre[keys[pred]]
            DefaultPert[:, pred] = DUncertainties_pre[keys[pred]]
            
            # Apply perturbations
            PMs2 = PMs + Pert
            DMs2 = PMs + DefaultPert
        
            # Update PMs3 and DMs3 dictionaries
            for col, name in enumerate(names):
                PMs3[name] = PMs2[:, col].tolist()
                DMs3[name] = DMs2[:, col].tolist()
        
            # Run preprocess_applynets for perturbed and default data
            VTF = False
            Eqs2, n2, VerbTF2, EstDates2, C2, PerKgSwTF2, MeasUncerts2 = defaults(
                var,
                PMs2,
                OutputCoordinates,
                Equations=Equations,
                EstDates=EstDates,
                VerboseTF = VTF,
            )
            U_pre2, DU_pre2 = measurement_uncertainty_defaults(
                n2,
                PMs3,
                MeasUncerts2
            )
            InputAll2 = inputdata_organize(
                EstDates2,
                C2,
                PMs3,
                U_pre2
            )
            InputAll3 = inputdata_organize(   
                EstDates2,
                C2,
                DMs3,
                U_pre2
            )
            PMs3, InputAll2 = temperature_define(
                var,
                PMs3,
                InputAll2,
                **kwargs
            )
            DMs3, InputAll3 = temperature_define(
                var,
                DMs3,
                InputAll3,
                **kwargs
            )
            code2, unc_combo_dict2, dunc_combo_dict2 = iterations(
                var,
                Eqs2,
                PerKgSwTF2,
                C2,  
                PMs3, 
                InputAll2,
                U_pre2,
                DU_pre2   
            )
            code3, unc_combo_dict3, dunc_combo_dict3 = iterations(
                var,  
                Eqs2,
                PerKgSwTF2,
                C2, 
                DMs3,
                InputAll3,
                U_pre2, 
                DU_pre2
            )
            df2 = define_polygons(C2)
            EstAtl2, EstOther2 = run_nets(
                var,
                Eqs2,   
                code2
            )
            EstAtl3, EstOther3 = run_nets(
                var, 
                Eqs2,
                code3
            )
            PertEst = process_netresults(
                Eqs2,  
                code2,    
                df2,
                EstAtl2,
                EstOther2
            )
            DefaultPertEst = process_netresults(
                Eqs2,
                code3,
                df2,
                EstAtl3,
                EstOther3
            )
            
            # Extract estimates and perturbation results
            estimates = [np.array(v) for v in Estimate.values()]
            pertests = [np.array(v) for v in PertEst.values()]
            defaultpertests = [np.array(v) for v in DefaultPertEst.values()]
             
            # Initialize result lists
            PertDiff, DefaultPertDiff, Unc_sub2, DUnc_sub2 = [], [], [], []

            for c in range(len(Equations)):
            # Compute differences and squared differences using numpy
                PD = estimates[c] - pertests[c]
                DPD = estimates[c] - defaultpertests[c]
                Unc_sub1 = (estimates[c] - pertests[c])**2
                DUnc_sub1 = (estimates[c] - defaultpertests[c])**2
                
               # Append results
                PertDiff.append(PD)
                DefaultPertDiff.append(DPD)
                Unc_sub2.append(Unc_sub1)
                DUnc_sub2.append(DUnc_sub1)
            Pertdv.append(PertDiff)
            DPertdv.append(DefaultPertDiff)
            Unc.append(Unc_sub2)
            DUnc.append(DUnc_sub2)
        PD_final.append(Pertdv)
        DPD_final.append(DPertdv)
        Unc_final.append(Unc)
        DUnc_final.append(DUnc)
            
    # Compute final uncertainty propagation
            
    est = [np.array(v) for v in Estimate.values()]
    emlr_combined = {k: v for d in emlr for k, v in d.items()}
    Uncertainties = {}
    for dv in range(0, len(DesiredVariables)):
        dvu = []
        for eq in range(0, len(Equations)):
            sumu = []
            name = DesiredVariables[dv] + str(Equations[eq])
            for n in range(0, len(est[0])):
                # Collect uncertainty contributions from each perturbation
                u =  np.array([Unc_final[dv][pre][eq][n] for pre in range(len(PredictorMeasurements))])
                du = np.array([DUnc_final[dv][pre][eq][n] for pre in range(len(PredictorMeasurements))])
                eu = emlr_combined[name][n]
                # Final uncertainty formula
                total_uncertainty = np.sqrt(np.sum(u) - np.sum(du) + eu**2)
                sumu.append(total_uncertainty)
            dvu.append(sumu)

            Uncertainties[name] = sumu

    return Uncertainties




def temperature_define(
    DesiredVariables, 
    PredictorMeasurements, 
    InputAll, 
    **kwargs
):

    """ 
   A small function to define temperature as needed and adjust
        InputAll and PredictorMeasurements acccordingly.

    Inputs:
        DesiredVariables: List of desired output variables
        PredictorMeasurements: Dictionary of user input measurements for predictors
        InputAll: Dictionary of pre-adjusted and combined input and defined data
        **kwargs: Please see README for more information

    Ouputs:
        PredictorMeasurements: Dictionary of user input measurements, adjusted for temperature
            as needed
        InputAll: Combined and processed necessary data to run the LIR, adjusted now for
            temperature as needed
    """

    import numpy as np

    n = max(len(v) for v in PredictorMeasurements.values()) # Recalculating number of required estimates

    # Printing a custom warning if temperature is absent but needed
    if "EstDates" in kwargs and "pH" in DesiredVariables:
        if "temperature" not in PredictorMeasurements:
            print(
                "Warning: Carbonate system calculations will be used to adjust the pH, but no temperature is"
                "specified so 10 C will be assumed. If this is a poor estimate for your region, consider supplying"
                "your own value in the PredictorMeasurements input."
            )
            # Assuming temperature is 10 C if not defined
            Temperature = np.full(n, 10)
        else:
            # Reading in temperature if defined
            Temperature = np.array(InputAll["Temperature"])

        # Adjusting dictionaries for the updated temperature definition
        PredictorMeasurements["temperature"] = Temperature
        InputAll["temperature"] = Temperature

    return PredictorMeasurements, InputAll
    



def run_nets(DesiredVariables, Equations, code={}):

    """
    Running neural nets

    Inputs:
        DesiredVariables: List of variables for estimates
        Equations: List of desired equations
        code: Dictionary of preprocessed measurements

    Outputs:
        EstAtl: Dictionary of estimates for the Atlantic and Arctic
            Oceans
        EstOther: Dictionary of estimates for not Altnatic/Arctic
    """

    import importlib
    import numpy as np

    # Predefining dictionaries to populate
    EstAtl, EstOther = {}, {}
    P, Sd, Td, Ad, Bd, Cd = {}, {}, {}, {}, {}, {}

    # Calculating inputs for nets and formatting them
    for name, value in code.items():
        cosd = np.cos(np.deg2rad(value["Longitude"] - 20)).tolist()
        sind = np.sin(np.deg2rad(value["Longitude"] - 20)).tolist()
        lat, depth = value["Latitude"].tolist(), value["Depth"].tolist()
        # Convert columns to lists of floats
        Sd[name] = value["S"].astype(float).tolist()
        Td[name] = value["T"].astype(float).tolist()
        Ad[name] = value["A"].astype(float).tolist()
        Bd[name] = value["B"].astype(float).tolist()
        Cd[name] = value["C"].astype(float).tolist()

   # Define a mapping from equations to the list of variable dictionaries
    equation_map = {
        1: [Sd, Td, Ad, Bd, Cd],
        2: [Sd, Td, Ad, Cd],
        3: [Sd, Td, Bd, Cd],
        4: [Sd, Td, Cd],
        5: [Sd, Td, Ad, Bd],
        6: [Sd, Td, Ad],
        7: [Sd, Td, Bd],
        8: [Sd, Td],
        9: [Sd, Ad, Bd, Cd],
        10: [Sd, Ad, Cd],
        11: [Sd, Bd, Cd],
        12: [Sd, Cd],
        13: [Sd, Ad, Bd],
        14: [Sd, Ad],
        15: [Sd, Bd],
        16: [Sd]
    }

    # Create the correct vector for each equation case
    for e in Equations:
        for v in DesiredVariables:
            name = v + str(e)
            # Get the corresponding variables for the equation
            variables = [var[name] for var in equation_map[e]]
            P[name] = [[[cosd, sind, lat, depth] + variables]]
            netstimateAtl, netstimateOther = [], []
            for n in range(1, 5):   
                fOName = f"NeuralNetworks.ESPER_{v}_{e}_Other_{n}"
                fAName = f"NeuralNetworks.ESPER_{v}_{e}_Atl_{n}"
                moda = importlib.import_module(fAName)
                modo = importlib.import_module(fOName)
                from importlib import reload
                reload(moda)
                reload(modo)   
                # Running the nets
                netstimateAtl.append(moda.PyESPER_NN(P[name]))
                netstimateOther.append(modo.PyESPER_NN(P[name]))
        
            # Process estimates for Atlantic and Other regions
            EstAtlL = [[netstimateAtl[na][0][eatl] for na in range(4)] for eatl in range(len(netstimateAtl[0][0]))]
            EstOtherL = [[netstimateOther[no][0][eoth] for no in range(4)] for eoth in range(len(netstimateOther[0][0]))]
        
            # Store the result
            EstAtl[name] = EstAtlL
            EstOther[name] = EstOtherL
            
    return EstAtl, EstOther




def define_polygons(C={}):

    """
    Defining and structuring indexing within ocean region polygons.
    First defines the polygons, then assesses the location of user-provided
    coordinates within polygons.

    Inputs:
        C: Dictionary of adjusted coordinates

    Output:
        df: Dictionary of adjusted coordinates with boolean indicators for specific
            ocean regions
    """

    import numpy as np
    import matplotlib.path as mpltPath

    # Define polygons for Atlantic and Arctic (AA) or other (Else) ocean basins
    LNAPoly = np.array([[300, 0], [260, 20], [240, 67], [260, 40], [361, 40], [361, 0], [298, 0]])
    LSAPoly = np.array([[298, 0], [292, -40.01], [361, -40.01], [361, 0], [298, 0]])
    LNAPolyExtra = np.array([[-1, 50], [40, 50], [40, 0], [-1, 0], [-1, 50]])
    LSAPolyExtra = np.array([[-1, 0], [20, 0], [20, -40], [-1, -40], [-1, 0]])
    LNOPoly = np.array([[361, 40], [361, 91], [-1, 91], [-1, 50], [40, 50], [40, 40], [104, 40], [104, 67], [240, 67],
                    [280, 40], [361, 40]])
    xtra = np.array([[0.5, -39.9], [.99, -39.9], [0.99, -40.001], [0.5, -40.001]])
    polygons = [LNAPoly, LSAPoly, LNAPolyExtra, LSAPolyExtra, LNOPoly, xtra]

    # Create Paths
    paths = [mpltPath.Path(poly) for poly in polygons]

    # Extract coordinates
    longitude, latitude, depth = np.array(C["longitude"]), np.array(C["latitude"]), np.array(C["depth"])

    # Check if coordinates are within each polygon
    conditions = [path.contains_points(np.column_stack((longitude, latitude))) for path in paths]

    # Combine conditions
    AAIndsM = np.logical_or.reduce(conditions)

    # Adding Bering Sea, S. Atl., and S. African Polygons separately
    Bering = np.array([[173, 70], [210, 70], [210, 62.5], [173, 62.5], [173, 70]])
    beringpath = mpltPath.Path(Bering)
    beringconditions = beringpath.contains_points(np.column_stack((longitude, latitude)))
    SAtlInds, SoAfrInds = [], []
    for i, z in zip(longitude, latitude):
        # Check if the conditions are met for Southern Atlantic
        if (-34 > z > -44):  # Check latitude first to reduce unnecessary checks
            if i > 290 or i < 20:
                SAtlInds.append('True')   
            else:
                SAtlInds.append('False')

            # Check if the condition is met for Southern Africa
            if 19 < i < 27:
                SoAfrInds.append('True')
            else:
                SoAfrInds.append('False')
        else:
            SAtlInds.append('False')
            SoAfrInds.append('False')

    # Create Dictionary with boolean indicators
    df = {'AAInds': AAIndsM, 'BeringInds': beringconditions, 'SAtlInds': SAtlInds, \
        'SoAfrInds': SoAfrInds, 'Lat': latitude, 'Lon': longitude, 'Depth': depth}
    
    return df




def mixed(DesiredVariables, Path, OutputCoordinates={}, PredictorMeasurements={}, **kwargs):
    
    """
    Python interpretation of ESPER_Mixedv1.1

    Empirical Seawater Property Estimation Routines: Estimates seawater properties and estimate uncertainty from combinations of other parameter
    measurements.  PYESPER_Mixed refers specifically to code that averages estunated from PyESPER_NN and PyESPER_LIR. See either subfunction for
    comments. The input arguments are the same for this function and for both subfunctions.
    
    *************************************************************************
    Please send questions or related requests about PyESPER to lmdias@uw.edu.
    ************************************************************************* 
    """
    import time
    import numpy as np
    from .lir import lir
    from .nn import nn
    
    tic = time.perf_counter()

    # Fetch estimates and uncertainties from PyESPER_LIR and PyESPER_NN
    EstimatesLIR, _, UncertaintiesLIR = lir(DesiredVariables, Path, OutputCoordinates, PredictorMeasurements, **kwargs)
    EstimatesNN, UncertaintiesNN = nn(DesiredVariables, Path, OutputCoordinates, PredictorMeasurements, **kwargs)

    Estimates, Uncertainties = {}, {}
    for est_type in EstimatesLIR.keys():
        estimates_lir = np.array(EstimatesLIR[est_type]).flatten() # need to flatten to match nn output shape
        estimates_nn = np.array(EstimatesNN[est_type]).astype(np.float64) # DICx - where x = equation number - returns string values which prevents proper averaging with LIR estimates
        uncertainties_lir = np.array(UncertaintiesLIR[est_type])
        uncertainties_nn = np.array(UncertaintiesNN[est_type])
        Estimates[est_type] = np.mean([estimates_lir, estimates_nn], axis=0).tolist()
        Uncertainties[est_type] = np.minimum(uncertainties_lir, uncertainties_nn).tolist()

    toc = time.perf_counter()
    print(f"PyESPER_Mixed took {toc - tic:0.4f} seconds, or {(toc-tic)/60:0.4f} minutes to run")    

    return Estimates, Uncertainties



def defaults (DesiredVariables, PredictorMeasurements={}, OutputCoordinates={}, **kwargs):

    """
    Set default values and bookkeep inputs.

    Inputs: 
        DesiredVariables: List of desired output variables (user-requested)
        PredictorMeasurements: Dictionary of user-provided predictor mesasurements (salinity, etc.)
        OutputCoordinates: Dictionary of user-provided coordinates
        **kwargs: Please see README for more information

    Outputs:
        Equations: numpy array of equations (either user-defined or default)
        n: Scalar representing number of required estimates for each variable-equation comination
        e: Scalar representing number of requested equations
        p: Scalar representing number of requested output variables
        VerboseTF: Boolean read-in of whether user wants to suppress optional warnings
        C: Dictionary of processed geographic coordinates
        PerKgSwTF: Boolean representing whether user input is in molal or molar units
        MeasUncerts: Dictionary of user input measurement uncertainty values or empty 
            dictionary if not provided   
    """

    import numpy as np

    # Check and define Equations based on user-defined kwargs, or use default values
    Equations = kwargs.get("Equations", list(range(1, 17)))
    
    # Reading dimensions of user input
    n = max(len(v) for v in OutputCoordinates.values()) 
                
    # Checking kwargs for presence of VerboseTF and EstDates, and Equations, and defining defaults, as needed
    VerboseTF = kwargs.get("VerboseTF", True)
        
    # Set EstDates based on kwargs, defaulting to 2002.0 if not provided
    if "EstDates" in kwargs:
        d = np.array(kwargs["EstDates"])
        if len(d) != n:
            EstDates = np.tile(d, (n + 1, 1)).reshape(-1)
        else:
            EstDates = d
    else:
        EstDates = np.full(n, 2002.0)
        
    # Bookkeeping coordinates
    C = {}
    longitude = np.array(OutputCoordinates["longitude"])
    longitude[longitude > 360] = np.remainder(longitude[longitude > 360], 360)
    longitude[longitude < 0] = longitude[longitude<0] + 360
    C["longitude"] = longitude
    C["latitude"] = OutputCoordinates["latitude"]
    C["depth"] = OutputCoordinates["depth"]   
    
    # Defining or reading in PerKgSwTF
    PerKgSwTF = kwargs.get("PerKgSwTF", True)

    # Defining Measurement Uncertainties
    MeasUncerts = kwargs.get("MeasUncerts", {})

    # Validate MeasUncerts dimensions
    if MeasUncerts:
        if max(len(v) for v in MeasUncerts.values()) != n:
            if min(len(v) for v in MeasUncerts.values()) != 1:
                raise CustomError(
                    "MeasUncerts must be undefined, a vector with the same number of elements as "
                    "PredictorMeasurements has columns, or a matrix of identical dimension to PredictorMeasurements."
                )
        if len(MeasUncerts) != len(PredictorMeasurements):
            print("Warning: Different numbers of input uncertainties and input measurements.")

    return Equations, n, VerboseTF, EstDates, C, PerKgSwTF, MeasUncerts



def adjust_pH_DIC(DesiredVariables, VerboseTF, Dates, Est_pre={}, PredictorMeasurements={}, OutputCoordinates={}, **kwargs):

    """
    If present, adjusting pH and DIC for anthropogenic carbon (Cant) within LIRs. Cant adjustment methods
        are based on those from ESPERv1, which is a TTD-based assumption/simplification but does not
        use the Tracer-based Rapid Anthropogenic Carbon Estimation data product, TRACE. Rather,
        interpolation from a gridded product is used to produce estimates for the year 2002 and data is
        adjusted to/from this reference year. This is the first of three steps for Cant adjustment

    Inputs:
        DesiredVariables: List of desired variables to estimate
        VerboseTF: Boolean indicating whether the user wants suppression of warnings
        Dates: List of dates for estimates
        Est_pre: Dictionary of preliminary estimates for each variable-equation case scenario
        PredictorMeasurements: Dictionary of input measurements for each variable-equation case scenario
        OutputCoordinates: Dictionary of coordinates for locations of estimates
        **kwargs: Please see README for full informations

    Outputs:
        Cant_adjusted: Dictionary of values adjusted for anthropogenic carbon for each combination
        Cant: Numpy array of estimates for anthropogenic carbon for each estimate
        Cant2002: Numpy array of estimates for anthropogenic carbon in the year 2002 for each estimate
    """

    import numpy as np
    from PyESPER.simplecantestimatelr import simplecantestimatelr

    # Predefining output dictionary and formatting estimates
    Cant_adjusted={}
    combos2 = list(Est_pre.keys())
    values2 = []
    for c, v in Est_pre.items():
        vals = np.array([v])
        vals = vals.flatten()
        values2.append(vals)
    values2 = np.array(values2)

    # Predefining anthropogenic carbon numpy arrays
    n = len(Dates)
    Cant, Cant2002 = np.zeros(n), np.zeros(n)

    # Only proceed if adjustment is needed
    if "EstDates" in kwargs and ("DIC" in DesiredVariables or "pH" in DesiredVariables):
        if VerboseTF:
            print("Estimating anthropogenic carbon for PyESPER_LIR.")

        # Normalize longitude to [0, 360]
        longitude = np.mod(np.array(OutputCoordinates["longitude"]), 360)
        latitude = np.array(OutputCoordinates["latitude"])
        depth = np.array(OutputCoordinates["depth"])
    
        # Estimate anthropogenic carbon (Cant) and anthropogenic carbon for the year 2002 (Cant2002)
        Cant, Cant2002 = simplecantestimatelr(Dates, longitude, latitude, depth)
        Cant, Cant2002 = np.array(Cant), np.array(Cant2002)
    
        for combo in range(0, len(combos2)):
            comb = combos2[combo]
            val = values2[combo]
            est1 = []
        
            # Only adjust if combo is DIC
            if "dic" in comb.lower():
                adjusted = np.where(val == "nan", np.nan, val + Cant - Cant2002)
                est1.append(adjusted)
    
            if "dic" not in comb.lower():
                nanfix = np.where(val == "nan", np.nan, val)
                est1.append(nanfix)
        
            Cant_adjusted[combos2[combo]] = est1

    return Cant_adjusted, Cant, Cant2002 




def nn(DesiredVariables, Path, OutputCoordinates={}, PredictorMeasurements={}, **kwargs):

    """
    Neural networks for seawater property estimation as part of PyESPERsv1.0.0

    Inputs: 
        DesiredVariables: list of desired variables to estimate
        Path: User-defined computer path
        OutputCoordinates: List of coordinates to produce estimates for
        PredictorMeasurements: List of predictor measurements to use in NNs
        **kwargs: Optional inputs specific to users (please see README for full description)

    Outputs:
        Estimates: Dictionary of estimates for each equation-desired variable combination
        Uncertainties: Dictionary of uncertainties for each equation-desired variable combination
    """

    import time
    from PyESPER.errors import errors
    from PyESPER.defaults import defaults
    from PyESPER.lir_uncertainties import measurement_uncertainty_defaults
    from PyESPER.inputdata_organize import inputdata_organize
    from PyESPER.temperature_define import temperature_define
    from PyESPER.iterations import iterations
    from PyESPER.fetch_polys_NN import fetch_polys_NN
    from PyESPER.define_polygons import define_polygons
    from PyESPER.run_nets import run_nets
    from PyESPER.process_netresults import process_netresults
    from PyESPER.organize_nn_output import organize_nn_output
    from PyESPER.pH_DIC_nn_adjustment import pH_DIC_nn_adjustment
    from PyESPER.final_formatting import final_formatting

    # Starting the timer 
    tic = time.perf_counter()
    
    # Function that provides custom error messages for erroneous inputs
    errors(OutputCoordinates, PredictorMeasurements)

    # Function which calculates default measurement uncertainties
    Equations, n, VerboseTF, EstDates, C, PerKgSwTF, MeasUncerts = defaults(
        DesiredVariables,
        PredictorMeasurements,
        OutputCoordinates,
        **kwargs
    )

    # Function that processes the input values and default uncertainties and makes sense of it
    Uncertainties_pre, DUncertainties_pre = measurement_uncertainty_defaults(
        n,
        PredictorMeasurements,
        MeasUncerts
    )

    # Creating a dictionary of input data
    InputAll = inputdata_organize(
        EstDates,
        C,
        PredictorMeasurements,
        Uncertainties_pre
    )

    # Defining temperature if not provided in the correct format
    PredictorMeasurements, InputAll = temperature_define(
        DesiredVariables, 
        PredictorMeasurements,
        InputAll,
        **kwargs
    )

    # Iterating through possible variable-equation combinations 
    # to produce a usable dictionary of predictor values
    code, unc_combo_dict, dunc_combo_dict = iterations(
        DesiredVariables,
        Equations,
        PerKgSwTF,
        C,
        PredictorMeasurements,
        InputAll,
        Uncertainties_pre,
        DUncertainties_pre
    )

    # Creating boolean indicators for different ocean regions
    df = define_polygons(C)

    # Running the actual neural nets
    EstAtl, EstOther = run_nets(
        DesiredVariables, 
        Equations, 
        code
    )

    # Processing and organizing results from nets, including regional
    # smoothing based on boolean indicators
    Estimates = process_netresults(
        Equations, 
        code, 
        df, 
        EstAtl, 
        EstOther
    )

    # Organize output and iteratively calculate uncertainties
    Uncertainties = organize_nn_output(
        Path,
        DesiredVariables,
        OutputCoordinates,
        PredictorMeasurements,
        **kwargs
    )

    # Adjust pH and DIC for anthropogenic carbon
    YouHaveBeenWarnedCanth=False
    Cant_adjusted = pH_DIC_nn_adjustment(
        Path,
        DesiredVariables, 
        Estimates,
        YouHaveBeenWarnedCanth,
        OutputCoordinates,
        PredictorMeasurements,
        **kwargs
    )
    
    # Final formatting and presentation of code
    Estimates = final_formatting(
        DesiredVariables,
        Cant_adjusted,
        Estimates
    )
       
    toc = time.perf_counter()
    print(f"PyESPER_NN took {toc - tic:0.4f} seconds, or {(toc-tic)/60:0.4f} minutes to run")

    return Estimates, Uncertainties



def interpolate(Gdf={}, AAdata={}, Elsedata={}):

    """
    This LIR function performs the interpolation on user-defined data

    Inputs:
        Gdf: Dictionary of pre-trained data for ESPER v1 (processed)
        AAdata: Dictionary of user input for Atlantic or Arctic
        Elsedata: Dictionary of user input not for Atlantic/Arctic

    Outputs:
        aaLCs: List of points to be interpolated within the Atlantic or Arctic
            regions
        aaInterpolants_pre: Scipy interpolant for Atlantic/Arctic region
        elLCs: List of points to be inteprolated outside of Atlantic/Arctic
        elInterpolants_pre: Scipy interpolant for outside of Atlantic/Arctic
    """

    import numpy as np
    from scipy.spatial import Delaunay
    import scipy.interpolate

    # Obtain data from the dictionaries
    Gvalues = list(Gdf.values())
    AAOvalues, ElseOvalues = list(AAdata.values()), list(Elsedata.values())

    def process_grid(grid_values, data_values):

        """
        A functionto help process data from grid and user data for interpolations
            and interpolate based upon a Delaunay triangulation, using scipy's
            LinearNDInterpolator
        """

        results = []
        for i in range(len(grid_values)):
            grid = grid_values[i]
            points = np.array([list(grid['lon']), list(grid['lat']), list(grid['d2d'])]).T
            tri = Delaunay(points)

            values = np.array([
                list(grid['C_alpha']),
                list(grid['C_S']),
                list(grid['C_T']),
                list(grid['C_A']),
                list(grid['C_B']),
                list(grid['C_C'])
            ]).T
            interpolant = scipy.interpolate.LinearNDInterpolator(tri, values)

            data = data_values[i]
            points_to_interpolate = (list(data['Longitude']), list(data['Latitude']), list(data['d2d']))
            results.append(interpolant(points_to_interpolate))
           
        return results, interpolant
            
    # Process AA (Atlantic/Arctic) and Else grids
    aaLCs, aaInterpolants_pre = process_grid(Gvalues, AAOvalues)
    elLCs, elInterpolants_pre = process_grid(Gvalues, ElseOvalues)
        
    return aaLCs, aaInterpolants_pre, elLCs, elInterpolants_pre




def pH_adjcalc(DesiredVariables, VerboseTF, Est_pre={}, Cant_adjusted={}, **kwargs):

    """
    Calculating pH adjustment if units are not in molal format

    Inputs:
        DesiredVariables: List of variables requested
        VerboseTF: Boolean option to suppress warnings
        Est_pre: Dictionary of estimates thus far for each variable equation combination
        Cant_adjusted: Dictionary of estimates adjusted for anthropogenic carbon for each
            combination
        **kwargs: Please see README for full information

    Outputs:
        Cant_adjusted: Dictionary of estimates adjusted for anthropogenic carbon
        combos2: List of equation case-desired variable combinations
        values2: List of estimates
    """

    combos2 = list(Est_pre.keys())
    values2 = list(Est_pre.values())

    if kwargs.get("pHCalcTF") and "pH" in DesiredVariables:
        if VerboseTF:
            print("Recalculating the pH to be appropriate for pH values calculated from TA and DIC.")

        for combo in range(0, len(combos2)):
            if combos2[combo].startswith("pH"):
                pH_adjcalc_Est = []
                pH_adjcalc = values2[combo]
                for v in pH_adjcalc:
                    pH_adjcalc_Est.append((pH_adjcalc[v]+0.3168)/1.0404)
            Cant_adjusted[combos2[combo]] = pH_adjcalc_Est

    return Cant_adjusted, combos2, values2




def iterations(
    DesiredVariables,
    Equations,
    PerKgSwTF,
    C={},
    PredictorMeasurements={},
    InputAll={},
    Uncertainties={},
    DUncertainties={}
):

    """
    A function to iterate and define equation iputs depending upon 
    Equation, DesiredVariable, and other user specifications.

    Inputs:
        DesiredVariables: List of desired variables for estimates
        Equations: List of equations to be used for estimates
        PerKgSwTF: Boolean indicating whether user input was in molal (default) or molar units
        C: Dictionary of pre-processed user geographic coordinates
        PredictorMeasurements: Dictionary of all input, with temperature processed as needed
        InputAll: Dictionary of deafault and user inputs with order stamp
        Uncertainties: Dictionary of user-defined uncertainties (or default if missing)
        DUncertainties: Dictionary of default uncertainties

    Output:
        code: Dictionary of filled-in combinations of predictor measurements relative to the reques$
            variable combination
        unc_combo_dict: Dictionary of filled-in uncertainty combinations for the requisite equation$
        dunc_combo_dict: Dictionary of filled-in default uncertainty combinations, as above

    NOTE: This function uses the now-deprecated seawater package for some calculations. This aligns with the
        current version of ESPERs (v1.0); however, this will be updated to the Gibb's seawater package in the
        next release.
    """

    import numpy as np
    import seawater as sw

    n = max(len(v) for v in C.values()) # Once again calculate number of estimates requested

    # Creating numpy arrays for depth, latitude, and salinity
    depth, latitude, salinity = np.array(C["depth"]), np.array(C["latitude"]), np.array(PredictorMeasurements["salinity"])

    # Calculating potential temperature (will be updated to conservative temperature in the next release)
    temp = np.array(PredictorMeasurements["temperature"]) if "temperature" in PredictorMeasurements else np.full(n, 10)
    temp_sw = sw.ptmp(salinity, temp, sw.pres(depth, latitude), pr=0)
    # Some whole numbers are treated incorrectly as strings by Python, this alleviates this problem
    temp_map = {3: 3.000000001, 4: 4.000000001, 5: 5.000000001, 6: 6.000000001}

    # Convert temp_sw to a NumPy array
    temp_sw_np = np.array(temp_sw)

    # Addressing potentially erroneous values
    mapped_temp = np.vectorize(lambda t: temp_map.get(t, 10 if t < -100 else t))(temp_sw_np)

    # Format each temperature to a string using 15 significant digits, necessary to propertly
    # process iterations in the next step
    temperature_processed = np.array(["{:.15g}".format(val) for val in mapped_temp])

    # Calculating AOU from oxygen and defining nan's as applicable
    if "oxygen" in PredictorMeasurements:
        oxyg = np.array(PredictorMeasurements["oxygen"])
        oxyg_sw = sw.satO2(salinity, temp_sw)*44.6596 - (oxyg)
    else:
        oxyg_sw = np.tile("nan", n)
    # For very small AOUs Python processes the numbers incorrectly, this rounds those to zero to avoid this
    for i in range(len(oxyg_sw)):
        if oxyg_sw[i] != "nan" and -0.0001 < oxyg_sw[i] < 0.0001:
            oxyg_sw[i] = 0
    # Format AOU to string as with temperature, for processing within iterations
    oxygen_processed = ["{:.5g}".format(o) if o != "nan" else o for o in oxyg_sw]
   
    # Process predictor measurements as np arrays or nan's, as applicable
    processed_measurements = {}
    for param in ["phosphate", "nitrate", "silicate"]:
        processed_measurements[param] = (
    np.array(PredictorMeasurements[param]) if param in PredictorMeasurements else np.tile("nan", n)
        )
    
    phosphate_processed = processed_measurements["phosphate"]
    nitrate_processed = processed_measurements["nitrate"]
    silicate_processed = processed_measurements["silicate"]
            
    # Adjust molar to molal units as needed
    if not PerKgSwTF:
        densities = sw.dens(salinity, temperature_processed, sw.pres(depth, latitude)) / 1000
        for nutrient in ["phosphate", "nitrate", "silicate"]:
            if nutrient in PredictorMeasurements:
                globals()[f"{nutrient}_processed"] /= densities
            
    # Convert equations to string for processing
    EqsString = [str(e) for e in Equations]

    # Definition of which input variable is needed for which property
    NeededForProperty = {
            "TA": [1, 2, 4, 6, 5],
            "DIC": [1, 2, 4, 6, 5],
            "pH": [1, 2, 4, 6, 5],
            "phosphate": [1, 2, 4, 6, 5],
            "nitrate": [1, 2, 3, 6, 5],
            "silicate": [1, 2, 3, 6, 4], 
            "oxygen": [1, 2, 3, 4, 5]
        }
    
    # Definition of equations
    VarVec = {
            "1": [1, 1, 1, 1, 1],
            "2": [1, 1, 1, 0, 1],
            "3": [1, 1, 0, 1, 1],
            "4": [1, 1, 0, 0, 1],
            "5": [1, 1, 1, 1, 0],
            "6": [1, 1, 1, 0, 0],
            "7": [1, 1, 0, 1, 0],
            "8": [1, 1, 0, 0, 0],
            "9": [1, 0, 1, 1, 1],
            "10": [1, 0, 1, 0, 1],
            "11": [1, 0, 0, 1, 1],
            "12": [1, 0, 0, 0, 1],
            "13": [1, 0, 1, 1, 0],
            "14": [1, 0, 1, 0, 0],
            "15": [1, 0, 0, 1, 0], 
            "16": [1, 0, 0, 0, 0], 
        }

    # Pre-defining lists and dictionaries to be used for output
    product, product_processed, name = [], [], []
    need, precode = {}, {}
         
    # Create a list of names and process products
    replacement_map = {
        "0": "nan",
        "1": "salinity",
        "2": "temperature",
        "3": "phosphate",
        "4": "nitrate",
        "5": "silicate",
        "6": "oxygen"
    }
            
    # Determining exactly which variable-equation scenarios required
    for d in DesiredVariables:
        dv = np.array(NeededForProperty[d])
        for e in EqsString:
            eq = np.array(VarVec[e])
            prename = d + e
            name.append(prename)  
            
            product.append(eq * dv)
            prodnp = np.array(eq * dv)
    
            # Replace values using the mapping   
            processed = np.vectorize(lambda x: replacement_map.get(str(x), x))(prodnp)
            need[prename] = processed
    
    for p in range(0, len(product)): # Same but for list of input values
        prodnptile = np.tile(product[p], (n, 1))
        prodnptile = prodnptile.astype("str")
        
        for v in range(0, len(salinity)):
            prodnptile[v][prodnptile[v] == "0"] = "nan"
            prodnptile[v][prodnptile[v] == "1"] = salinity[v]
            prodnptile[v][prodnptile[v] == "2"] = temperature_processed[v]
            prodnptile[v][prodnptile[v] == "3"] = phosphate_processed[v]
            prodnptile[v][prodnptile[v] == "4"] = nitrate_processed[v]
            prodnptile[v][prodnptile[v] == "5"] = silicate_processed[v]
            prodnptile[v][prodnptile[v] == "6"] = oxygen_processed[v]
            product_processed.append(prodnptile)
        
    listofprods = np.arange(0, len(product) * n, n)
    prodlist = []
            
    names_values = list(need.values())
    names_keys = list(need.keys()) 
            
    # Repeat for uncertainties
    unc_combo_dict = {}
    dunc_combo_dict = {}
            
    for numb_combos, names_keyscombo in enumerate(names_values):
        # A function to help process the uncertainties and nan's
        def define_unc_arrays(lengthofn, listorder, parnames, unames):
            for numoptions in range(0, len(parnames)):
                if names_keyscombo[listorder] == parnames[numoptions]:
                    udfvalues = np.array(Uncertainties[unames[numoptions]])
                    dudfvalues = np.array(DUncertainties[unames[numoptions]])
                elif names_keyscombo[listorder] == "nan":
                    udfvalues = np.empty((lengthofn))
                    udfvalues[:] = np.nan
                    dudfvalues = np.empty((lengthofn))
                    dudfvalues[:] = np.nan
            return udfvalues, dudfvalues

        for names_items in range(0, len(names_keyscombo)):
            udfvalues1 = np.array(Uncertainties['sal_u'])
            dudfvalues1 = np.array(DUncertainties['sal_u'])
            udfvalues2, dudfvalues2 = define_unc_arrays(n, 1, ["temperature"], ["temp_u"])
            udfvalues3, dudfvalues3 = define_unc_arrays(n, 2, ["nitrate", "phosphate"], ["nitrate_u", "phosphate_u"])
            udfvalues4, dudfvalues4 = define_unc_arrays(n, 3, ["oxygen", "nitrate"], ["oxygen_u", "nitrate_u"])
            udfvalues5, dudfvalues5 = define_unc_arrays(n, 4, ["silicate", "nitrate"], ["silicate_u", "nitrate_u"])
    
        # Convert to NumPy arrays for efficient comparison
        udfvalues = np.array([udfvalues1, udfvalues2, udfvalues3, udfvalues4, udfvalues5])
        dudfvalues = np.array([dudfvalues1, dudfvalues2, dudfvalues3, dudfvalues4, dudfvalues5])
    
        # Update `udfvalues` based on `dudfvalues` using element-wise maximum
        udfvalues = udfvalues.astype(np.float64)
        dudfvalues = dudfvalues.astype(np.float64)
        udfvalues = np.maximum(udfvalues, dudfvalues)
                    
        new_unames = ['US', 'UT', 'UA', 'UB', 'UC']
        # Transpose and convert to dictionaries of NumPy arrays
        uncertaintyvalues_dict = {name: np.array(col) for name, col in zip(new_unames, udfvalues)}
        duncertaintyvalues_dict = {name: np.array(col) for name, col in zip(new_unames, dudfvalues)}
                    
        # Update the target dictionaries  
        unc_combo_dict[names_keys[numb_combos]] = uncertaintyvalues_dict
        dunc_combo_dict[names_keys[numb_combos]] = duncertaintyvalues_dict
        
    # Append the required products to `prodlist` and populate `precode`
    prodlist = [product_processed[item] for item in listofprods]
    precode = {name[i]: prodlist[i] for i in range(len(listofprods))}
            
    S, T, A, B, Z, code = [], [], [], [], [], {}
            
    # Final filling in of values and code
    for value in precode.values():
        S.append(value[:, 0])
        T.append(value[:, 1])
        A.append(value[:, 2])
        B.append(value[:, 3])
        Z.append(value[:, 4])
        
    codenames = list(precode.keys())
                    
    code = {}
    for n, code_name in enumerate(codenames):
        # Create a dictionary for each set of data
        p = {
            "S": np.array(S[n]),
            "T": np.array(T[n]),
            "A": np.array(A[n]),
            "B": np.array(B[n]),
            "C": np.array(Z[n])
        }
    
        for col in ["Order", "Dates", "Longitude", "Latitude", "Depth", "Salinity_u", "Temperature_u", "Phosphate_u", "Nitrate_u", "Silicate_u", "Oxygen_u"]:
            p[col] = np.array(InputAll[col])

        # Assign the dictionary to the code dictionary
        code[code_name] = p

    return code, unc_combo_dict, dunc_combo_dict




def fetch_data (DesiredVariables, Path):
    
    """
    Gathers the necessary LIR files that were pre-trained in MATLAB ESPERs

    Inputs:
        DesiredVariables: List of desired output estimate variables
        Path: User-defined computer path of locations of files

    Outputs:
        LIR_data: List of dictionaries of LIR data
    """

    from scipy.io import loadmat
    import os
    import numpy as np

    # Predefine dictionaries of output
    AAIndsCs, GridCoords, Cs = {}, {}, {}

    # Load necessary files
    for v in DesiredVariables:
        fname1 = os.path.join(Path, f"Mat_fullgrid/LIR_files_{v}_fullCs1.mat")
        fname2 = os.path.join(Path, f"Mat_fullgrid/LIR_files_{v}_fullCs2.mat")
        fname3 = os.path.join(Path, f"Mat_fullgrid/LIR_files_{v}_fullCs3.mat")
        fname4 = os.path.join(Path, f"Mat_fullgrid/LIR_files_{v}_fullGrids.mat")

        Cs1 = loadmat(fname1)
        Cs2 = loadmat(fname2)
        Cs3 = loadmat(fname3)
        Grid = loadmat(fname4)

        # Extract and store all arrays
        UncGrid = np.array(Grid["UncGrid"][0][0])
        GridCoodata = np.array(Grid["GridCoords"])
        AAInds = np.array(Grid["AAIndsM"])

        Csdata1 = np.array(Cs1["Cs1"])
        Csdata2 = np.array(Cs2["Cs2"])
        Csdata3 = np.array(Cs3["Cs3"])

        # Store as NumPy arrays
        AAIndsCs[v] = AAInds
        GridCoords[v] = GridCoodata

        # Combine along axis 1, then store each layer in list
        Csdata = np.concatenate((Csdata1, Csdata2, Csdata3), axis=1)
        Cs[v] = [Csdata[:, :, i] for i in range(Csdata.shape[2])]

    # Store all in one list
    LIR_data = [GridCoords, Cs, AAIndsCs, UncGrid]

    return LIR_data




from .final_formatting import final_formatting
from .organize_nn_output import organize_nn_output
from .adjust_pH_DIC import adjust_pH_DIC
from .input_AAinds import input_AAinds	
from .pH_adjcalc import pH_adjcalc
from .coefs_AAinds import coefs_AAinds
from .inputdata_organize import inputdata_organize	
from .pH_adjustment import pH_adjustment
from .defaults import defaults
from .interpolate import interpolate
from .pH_DIC_nn_adjustment import pH_DIC_nn_adjustment
from .define_polygons import define_polygons
from .iterations import iterations
from .process_netresults import process_netresults
from .emlr_estimate import emlr_estimate
from .lir_uncertainties import process_uncertainties
from .lir_uncertainties import measurement_uncertainty_defaults	
from .run_nets import run_nets
from .emlr_nn import emlr_nn			
from .simplecantestimatelr import simplecantestimatelr
from .errors import errors
from .temperature_define import temperature_define
from .fetch_data import fetch_data
from .fetch_polys_NN import fetch_polys_NN	
from .organize_data import organize_data
from .lir import lir
from .nn import nn
from .mixed import mixed



def coefs_AAinds(Equations, LIR_data):

    """
    Separates coefficients from MATLAB ESPERv1 into Atlantic and Arctic or other regions.

    Inputs:
        Equations: List of equations for use in ESPERs
        LIR_data: List of dictionaries of data from MATLAB trainings (pre-processed)

    Outputs:
        Gdf: Dictionary of pre-trained and processed LIR data for grid of coordinates
        CsDesired: Dictionary of equation coefficients based on user-defined output
    """

    import numpy as np

    # Use boolean for AA or Else to separate coefficients into Atlantic or not
    GridCoords, Cs, AAInds = LIR_data[:3]
    DVs, CsVs = list(Cs.keys()), list(Cs.values())
    ListVars = np.arange(len(AAInds))
    GridValues = np.array(list(GridCoords.values())[0], dtype=float)
    AAIndValues = np.array(list(AAInds.values())[0], dtype=float)

    lon_grid, lat_grid, d2d_grid, aainds = np.array((GridValues[:,0])), np.array((GridValues[:,1])), \
       np.array(GridValues[:,2])/25, np.array(AAIndValues[:,0])

    names = ['lon', 'lat', 'd2d', "C_alpha", "C_S", "C_T", "C_A", "C_B", "C_C", 'AAInds']
    Gdf, CsDesired = {}, {}

    # Moving data into pre-defined dictionaries
    for lvar, name in zip(ListVars, DVs):
        Cs2 = CsVs[:][lvar][:]
        for e in Equations:
            CsName = f'Cs{name}{e}'
            CsDesired[CsName] = Cs2[e-1][:]
            Cs3 = Cs2[e-1][:]
            C_alpha, C_S, C_T, C_A, C_B, C_C = np.array(Cs3[:,0]), np.array(Cs3[:,1]), np.array(Cs3[:,2]), np.array(Cs3[:,3]), \
                np.array(Cs3[:,4]), np.array(Cs3[:,5])
            Gdf[f"{name}{e}"] = {
                names[0]: np.array(lon_grid),
                names[1]: np.array(lat_grid),
                names[2]: np.array(d2d_grid),
                names[3]: np.array(C_alpha),
                names[4]: np.array(C_S),
                names[5]: np.array(C_T),
                names[6]: np.array(C_A),
                names[7]: np.array(C_B),
                names[8]: np.array(C_C),
                names[9]: np.array(aainds)
            }

    return Gdf, CsDesired




def inputdata_organize(
    EstDates, 
    C={}, 
    PredictorMeasurements={}, 
    Uncertainties={} 
):

    """
  This function preprocesses data into a dictionary for easier referencing

    Inputs:
        EstDates: List of preprocessed dates
        C: Dictionary of preprocessed geographic coordinates
        PredictorMeasurements: Dictionary of preprocessed inputs
        Uncertainties: Dictionary of preprocessed uncertainties

    Ouputs:
        InputAll: Dictionary of preprocessed coordinates, measurements, uncertainties, dates, and an indexing term
    """

    import numpy as np

    n = max(len(v) for v in C.values()) # Simply recalculating number of rows out

    # Redefining and organizing all data thus far, and adding an order/indexing stamp
    order = np.arange(n)
    InputAll = {
        "Order": order,
        "Longitude": np.array(C["longitude"]),
        "Latitude": np.array(C["latitude"]),
        "Depth": np.array(C["depth"]),
        "Salinity": PredictorMeasurements["salinity"],
        "Dates": EstDates,
        "Salinity_u": Uncertainties["sal_u"],
        "Temperature_u": Uncertainties["temp_u"],
        "Phosphate_u": Uncertainties["phosphate_u"],
        "Nitrate_u": Uncertainties["nitrate_u"],
        "Silicate_u": Uncertainties["silicate_u"],
        "Oxygen_u": Uncertainties["oxygen_u"]
    }

   # Map PredictorMeasurements keys to InputAll keys
    for key, label in {
        "temperature": "Temperature",
        "phosphate": "Phosphate",
        "nitrate": "Nitrate",
        "silicate": "Silicate",
        "oxygen": "Oxygen"
    }.items():
        if key in PredictorMeasurements:
            InputAll[label] = np.array(PredictorMeasurements[key])

    return InputAll






def process_uncertainties(param, default_factor, MeasUncerts, PredictorMeasurements, n):
   
    """ 
  Helps proces uncertainties as needed by formatting any possible inputs, calculating
        outputs, and processing "nan" values

    Inputs:
        param: String name of uncertainty to be calculated
        default_factor: Scalar default uncertainty value
        MeasUncerts: Dictionary of measurement uncertainty values
        PredictorMeasurements: Dictionary of input measurements
        n: Scalar number of estimates

    Ouputs:
        result: Numpy array of user uncertainty-based uncertainties for each param
        dresult: Numpy array of default uncertainties for each param
    """
    
    import numpy as np

   # Determining whether the defined seawater property uncertainty has been used as input
    if param in MeasUncerts:
        # Obtain the predefined uncertainty if provided
        result = np.array(MeasUncerts.get(param))
        # Determining if unique uncertainties for each measurement were provided of if
        # only one uncertainty per property was given for all properties, and formatting
        # to the same length if so
        if len(result) < n:
            result = np.tile(result, n)
        # Formatting the naming convention of default uncertainties
        if param.replace('_u', '') in PredictorMeasurements:
            dresult = np.array([i * default_factor for i in PredictorMeasurements[param.replace('_u', '')]])
        else:
            dresult = result
    # Giving defaults in the case where user-provided uncertainties were not provided
    else:
        if param.replace('_u', '') in PredictorMeasurements:
            result = np.array([i * default_factor for i in PredictorMeasurements[param.replace('_u', '')]])
            dresult = result
        else:
            result = np.tile('nan', n)
            dresult = np.tile(0, n)
    return result, dresult

def measurement_uncertainty_defaults(n, PredictorMeasurements={}, MeasUncerts={}):
    
    """
    Inputs:
        n: Scalar number of estimates requested
        PredictorMeasurements: Dictionary of predictor measurements used for analyses
        MeasUncerts: User-provided measurement uncertainties or empty dictionary, if not provided
            
    Outputs:
        Uncertainties_pre: Dictionary of user-provided measurement uncertainties
        DUncertainties_pre: Dictionary of default measurement uncertainties
    """
    
    import numpy as np

    Uncertainties_pre, DUncertainties_pre = {}, {}
            
    # User-input salinity measurement uncertainties
    sal_u = np.array(MeasUncerts.get("sal_u", [0.003]))
    sal_u = np.tile(sal_u, n) if len(sal_u) < n else sal_u
    # Default salinity measurement uncertainties
    sal_defu = np.tile(0.003, n)
    
    # User-defined and default temperature measurement uncertainties
    temp_u = np.tile(np.array(MeasUncerts.get("temp_u", [0.003])), n) if "temp_u" in MeasUncerts or "temperature" in PredictorMeasurements else np.tile("nan", n)
    temp_defu = np.tile(0.003 if "temp_u" in MeasUncerts or "temperature" in PredictorMeasurements else 0, n)
    
    # Process other parameters
    parameters = {
        "phosphate_u": 0.02,
        "nitrate_u": 0.02,
        "silicate_u": 0.02,
        "oxygen_u": 0.01
    }
        
    # User process_uncertainties function to calculate defaults and user-defined uncertainties for each
    # parameter in a dictionary
    for param, factor in parameters.items():
        Uncertainties_pre[param], DUncertainties_pre[param] = process_uncertainties(
            param,
            factor,
            MeasUncerts,
            PredictorMeasurements,
            n
        )
    
    # Update MeasUncerts and DefaultUAll dictionary keys to include salinity, temperature, and all other properties
    meas_uncerts_keys = ["sal_u", "temp_u", *parameters.keys()]
    
    # Populating the dictionaries
    Uncertainties_pre.update(dict(zip(meas_uncerts_keys, [sal_u, temp_u, *Uncertainties_pre.values()])))
    DUncertainties_pre.update(dict(zip(meas_uncerts_keys, [sal_defu, temp_defu, *DUncertainties_pre.values()])))
    
    return Uncertainties_pre, DUncertainties_pre




def input_AAinds(C={}, code={}):

    """
    Separates user-defined inpus into Atlantic and Arctic regions or other
        regions, defined as in ESPERv1 for MATLAB.

    Inputs:
        C: Dictionary of pre-adjusted grid coordinates
        code: Dictionary of iterated equation-case scenario inputs for
            user-requested variable-equation cases

    Outputs:
        AAdata: Dictionary of code data separated for areas encompassed by the
            Atlantic and Arctic Oceans only
        Elsedata: Dictionary of code data separated for areas not encompassed by
            the Atlantic and Arctic Oceans
    """

    import numpy as np
    import matplotlib.path as mpltPath

    # Define Polygons that encompass the Atlantic and Arctic Oceans, geographically
    LNAPoly = np.array([[300, 0], [260, 20], [240, 67], [260, 40], [361, 40], [361, 0], [298, 0]])
    LSAPoly = np.array([[298, 0], [292, -40.01], [361, -40.01], [361, 0], [298, 0]])
    LNAPolyExtra = np.array([[-1, 50], [40, 50], [40, 0], [-1, 0], [-1, 50]])
    LSAPolyExtra = np.array([[-1, 0], [20, 0], [20, -40], [-1, -40], [-1, 0]])
    LNOPoly = np.array([[361, 40], [361, 91], [-1, 91], [-1, 50], [40, 50], [40, 40], [104, 40], [104, 67], [240, 67],
                        [280, 40], [361, 40]])
    xtra = np.array([[0.5, -39.9], [.99, -39.9], [0.99, -40.001], [0.5, -40.001]])

    polygons = [LNAPoly, LSAPoly, LNAPolyExtra, LSAPolyExtra, LNOPoly, xtra]

    # Create Paths
    paths = [mpltPath.Path(poly) for poly in polygons]

    # Extract coordinates
    longitude, latitude, depth = np.array(C["longitude"]), np.array(C["latitude"]), np.array(C["depth"])

    # Check if coordinates are within each polygon
    conditions = [path.contains_points(np.column_stack((longitude, latitude))) for path in paths]

    # Combine conditions
    AAIndsM = np.logical_or.reduce(conditions)
    AAIndsM = AAIndsM.astype(int)

    # Create dictionary
    df = {
        'AAInds': AAIndsM,
        'Lat': latitude,
        'Lon': longitude,
        'Depth': depth
    }
                        
    for df_code in code.values():
        df_code['AAInds'] = df['AAInds']
    
    # Initialize dictionaries for AA (Atlantic Arctic) and Else (not Atlantic Arctic) data
    AAdata = {}   
    Elsedata = {}

    # Iterate over each key in code
    for i in code:
        # Extract data arrays from the DataFrame
        data_arrays = np.array([code[i][key] for key in ['Depth', 'Latitude', 'Longitude', 'S', 'T', 'A', 'B', 'C',
                                                                'Order', 'Salinity_u', 'Temperature_u', 'Phosphate_u',
                                                                'Nitrate_u', 'Silicate_u', 'Oxygen_u', 'AAInds']])
    
        # Unpack arrays into separate variables
        depth, latitude, longitude, S, T, A, B, C, order, sal_u, temp_u, phos_u, nitr_u, sil_u, oxyg_u, aainds = [
            np.array(arr, dtype=float) for arr in data_arrays
        ]
    
        # Scale the depth values
        depth = depth / 25
        
        # Reshape data arrays to match the number of rows
        NumRows_out = len(longitude)
        reshaped_data = [arr.reshape(NumRows_out, 1) for arr in [depth, latitude, longitude, S, T, A, B, C, order, sal_u,
                                                                 temp_u, phos_u, nitr_u, sil_u, oxyg_u, aainds]]
        dep, lat, lon, sal, temp, avar, bvar, cvar, orde, salu, tempu, phosu, nitru, silu, oxygu, aai = reshaped_data
    
        # Combine the data into one array for further splitting
        InputBool = np.hstack(reshaped_data)
    
        # Define columns for the final dictionary
        columns = ['d2d', 'Latitude', 'Longitude', 'S', 'T', 'A', 'B', 'C', 'Order', 'Salinity_u', 'Temperature_u',
                   'Phosphate_u', 'Nitrate_u', 'Silicate_u', 'Oxygen_u', 'AAInds']
        NumCols_out = len(columns)
        
        # Function to filter arrays based on condition
        def split(arr, cond):
            return arr[cond]
        
        # Split the data into AA and Else data
        InputAA_01 = split(InputBool, InputBool[:, -1] == 1) 
        InputElse_01 = split(InputBool, InputBool[:, -1] == 0)

        # Reshape and create dictionaries for AA and Else
        AAInput = {
            col: InputAA_01.reshape(len(InputAA_01), NumCols_out)[:, i]
            for i, col in enumerate(columns)
        }
        ElseInput = {
            col: InputElse_01.reshape(len(InputElse_01), NumCols_out)[:, i]
            for i, col in enumerate(columns)
        }
        
        # Store the results in the dictionaries
        AAdata[i] = AAInput
        Elsedata[i] = ElseInput
        
    return AAdata, Elsedata




def simplecantestimatelr(EstDates, longitude, latitude, depth):

    """
    Simple estimate of contribution of anthropogenic carbon to pH and DIC estimates.

    Inputs:
        EstDates: List of dates for which estimates will be made
        longitude:

    Ouptuts:
        CantMeas: List of anthropogenic carbon estimates
        Cant2002: List of anthropogenic carbon estimates for 2002
    """

    import numpy as np
    import pandas as pd
    from scipy.interpolate import griddata

    # Load interpolation points and values
    CantIntPoints = pd.read_csv('SimpleCantEstimateLR_full.csv')
    pointsi = (
        CantIntPoints['Int_long'] * 0.25,
        CantIntPoints['Int_lat'],
        CantIntPoints['Int_depth'] * 0.025,
    )
    values = CantIntPoints['values']

    # Scale input coordinates
    pointso = (
        np.array(longitude) * 0.25,
        np.array(latitude),
        np.array(depth) * 0.025,
    )

    # Interpolate and compute Cant2002
    Cant2002 = griddata(pointsi, values, pointso, method='linear')

    # Adjust for estimation dates
    EstDates = np.asarray(EstDates)
    CantMeas = Cant2002 * np.exp(0.018989 * (EstDates - 2002))

    return CantMeas, Cant2002




def process_netresults(Equations, code={}, df={}, EstAtl={}, EstOther={}):

    """
    Regional smoothing and processing net outputs

    Inputs:
        Equations: List of equations to use
        code: Dictionary of input for each equation-variable scenario
        df: Dictionary of coordinates and boolean indicators for regions
        EstAtl: Dictionary of estimates for the Atlantic and Arctic
            for each combination
        EstOther: Dictionary of estimates for not Atlantic and Arctic,
            for each combination

    Output:
        Estimate: Dictionary of estimates for each combination
    """

    import numpy as np

    # Function to provide the mean of values when needed
    def process_estimates(estimates):
        keys = list(estimates.keys())
        values = list(estimates.values())
        result = {}
        for i, key in enumerate(keys):
            result[key] = [np.mean(values[i][v]) for v in range(len(values[0]))]
        return result

    Esta = process_estimates(EstAtl)
    Esto = process_estimates(EstOther)

    # Processing regionally in the Atlantic and Bering
    EstA, EstB, EB2, ESat, ESat2, ESaf, Estimate  = {}, {}, {}, {}, {}, {}, {}

    for i in code:
        code[i]["AAInds"] = df["AAInds"]
        code[i]["BeringInds"] = df["BeringInds"]
        code[i]["SAtlInds"] = df["SAtlInds"]
        code[i]["SoAfrInds"] = df["SoAfrInds"]

    for codename, codedata in code.items():
        # Defining empty lists for data from each regions
        Estatl, Estb, eb2, Estsat, esafr = [], [], [], [], []
        aainds, beringinds, satlinds, latitude, safrinds = (
            codedata[key] for key in ["AAInds", "BeringInds", "SAtlInds", "Latitude", "SoAfrInds"]
        )
        Estatl = [Esta[codename][i] if aa_ind else Esto[codename][i] for i, aa_ind in enumerate(aainds)]
        
        # Smoothing for the Atlantic
        for lenatl in range(0, len(Estatl)):
            repeated_values = (latitude[lenatl]-62.5)/7.5
            B = np.tile(repeated_values, (1, len(Equations)))
            C = Esta[codename][lenatl]
            B1 = C * B
            repeated_values2 = (70-latitude[lenatl])/7.5
            D = np.tile(repeated_values2, (1, len(Equations)))
            E = Esto[codename][lenatl]
            B2 = E * D
            Estb.append(B1[0][0] + B2[0][0])
    
        eb2 = [Estb[j] if b_ind else Estatl[j] for j, b_ind in enumerate(beringinds)]
        # Smoothing for the Bering
        for n in range(0, len(satlinds)):   
            repeated_values = (latitude[n]+44)/10
            F1 = Esta[codename][n]
            F = np.tile(repeated_values, (1, len(Equations)))
            G1 = F1 * F
            repeated_values2 = (-34-latitude[n])/10
            H1 = Esto[codename][n]
            H = np.tile(repeated_values2, (1, len(Equations)))
            G2 = H1 * H
            Estsat.append(G1[0][0] + G2[0][0])

        EstA[codename], EstB[codename], EB2[codename], ESat[codename] = Estatl, Estb, eb2, Estsat
        
    # Regional processing for S. Atlantic
        ESat2[codename] = [
            ESat[codename][i] if satlinds[i] == "True" else EB2[codename][i]
            for i in range(len(satlinds))
        ]
            
    # Regional processing for S. Africa
        for s in range(0, len(safrinds)):
            repeated_values = (27-df["Lon"][s])/8
            F1 = ESat2[codename][s]
            F = np.tile(repeated_values, (1, len(Equations)))
            G1 = F1*F
            repeated_values2 = (df["Lon"][s]-19)/8
            H1 = Esto[codename][s]
            H = np.tile(repeated_values2, (1, len(Equations)))
            G2 = H1 * H
            esafr.append(G1[0][0] + G2[0][0])
            
        ESaf[codename] = esafr

        Estimate[codename] = [
            ESaf[codename][i] if safrinds[i] == "True" else ESat2[codename][i]
            for i in range(len(safrinds))
        ]
        
    # Bookkeeping blanks back to NaN as needed
    Estimate = {k: ('NaN' if v == '' else v) for k, v in Estimate.items()}
            
    return Estimate



def final_formatting(DesiredVariables, Cant_adjusted={}, Est_pre={}):

    """
    Formatting the final data output for estimates

    Inputs:
        DesiredVariables: List of desired variables to estimate
        Cant_adjusted: Dictionary of estimates adjusted for anthropogenic
            carbon for each combination
        Est_pre: Dictionary of estimates for each combination

    Output:
        Estimates: Dictionary of estimates for each combination
    """

    # Conditional to whether anthropogenic carbon was needed or not
    if ("pH" or "DIC") in DesiredVariables:
        Estimates=Cant_adjusted
        print("anthropogenic carbon has been incorporated into some estimates")
    else:
        Estimates=Est_pre
        print("anthropogenic carbon is not considered for these estimates")

    return Estimates




def organize_data(
    aaLCs, 
    elLCs, 
    aaInterpolants_pre, 
    elInterpolants_pre, 
    Gdf={}, 
    AAdata={},
    Elsedata={}
    ):

    """
    Organize interpolation output into more usable formatting and compute estimates

    Inputs:
        aaLCs: List of coefficients for input data from Atlantic/Arctic regions
        elLCs: List of coefficients for input data not from Atlantic/Arctic
        aaInterpolants_pre: Scipy interpolant for Atlantic/Arctic
        elInterpolants_pre: Scipy interpolant for not Atlantic/Arctic
        Gdf: Dictionary of grid for interpolation, separated into regions
        AAdata: Dictionary of user input data for Atlantic/Arctic
        ElseData: Dictionary of user input data not for Atlantic/Arctic

    Outputs:
        Estimate: Dictionary of estimates for each equation-desired variable combination
        CoefficientsUsed: Dictionary of dictionaries of coefficients for each equation-
            desired variable combination
    """

    import numpy as np

    # Determine combinations
    Gkeys = list(Gdf.keys())

    AAOvalues, ElseOvalues = list(AAdata.values()), list(Elsedata.values())
    # Initialize lists for storing interpolated values
    aaIntCT2, aaIntCA2, aaIntCB2, aaIntCC2, aaTo2, aaAo2, aaBo2, aaCo2 = [[] for _ in range(8)]
    elIntCT2, elIntCA2, elIntCB2, elIntCC2, elTo2, elAo2, elBo2, elCo2 = [[] for _ in range(8)]
    aaInterpolants, elInterpolants = {}, {}

    # Separate out stored values
    for i in range(0, len(aaLCs)):
        aaIntalpha, elIntalpha = aaLCs[i][:, 0], elLCs[i][:, 0]
        aaIntCS, elIntCS = aaLCs[i][:, 1], elLCs[i][:, 1]
        aaIntCT, elIntCT = aaLCs[i][:, 2], elLCs[i][:, 2]
        aaIntCA, elIntCA = aaLCs[i][:, 3], elLCs[i][:, 3]
        aaIntCB, elIntCB = aaLCs[i][:, 4], elLCs[i][:, 4]
        aaIntCC, elIntCC = aaLCs[i][:, 5], elLCs[i][:, 5]

        # A function to process data in the correct manner and
        # proper handling of nan's
        def process_list(int_values, val_values):
            int_values = np.asarray(int_values, dtype=float)
            val_values = np.asarray(val_values, dtype=object)
    
            # Replace NaNs in int_values with 0
            int2 = np.where(np.isnan(int_values), 0, int_values)
    
            # Replace string "nan" and float NaN in val_values with 0
            val_cleaned = []
            for val in val_values:
                if isinstance(val, str) and val.lower() == "nan":
                    val_cleaned.append(0)
                elif isinstance(val, float) and np.isnan(val):
                    val_cleaned.append(0)
                else:
                    val_cleaned.append(val)
            val2 = np.array(val_cleaned, dtype=object)
        
            return int2, val2

        # Correcting equation instances from nan to zero 
        # for calculations
        key = Gkeys[i]
        is_key_1 = key[-1] == "1" and key[-2] != "1"
        is_key_2 = key[-1] == "2" and key[-2] != "1"
        is_key_3 = key[-1] == "3" and key[-2] != "1"
        is_key_4 = key[-1] == "4" and key[-2] != "1"
        is_key_5 = key[-1] == "5" and key[-2] != "1"
        is_key_6 = key[-1] == "6" and key[-2] != "1"
        is_key_7 = key[-1] == "7"
        is_key_8 = key[-1] == "8"
        is_key_9 = key[-1] == "9"
        is_key_10 = key[-1] == "0" and Gkeys[i][-2] == "1"
        is_key_11 = key[-1] == "1" and key[-2] == "1"
        is_key_12 = key[-1] == "2" and key[-2] == "1"
        is_key_13 = key[-1] == "3" and key[-2] == "1"
        is_key_14 = key[-1] == "4" and key[-2] == "1"
        is_key_15 = key[-1] == "5" and key[-2] == "1"
        is_key_16 = key[-1] == "6" and key[-2] == "1"
                
        # Atlantic and Arctic data equation processing
        aaDatao = AAOvalues[i]
        aaSo, aaTo, aaAo, aaBo, aaCo = aaDatao['S'], aaDatao['T'], aaDatao['A'], aaDatao['B'], aaDatao['C']
            
        # Determine which values to use
        if is_key_1:
            aaIntCT2, aaIntCA2, aaIntCB2, aaIntCC2 = aaIntCT, aaIntCA, aaIntCB, aaIntCC
            aaTo2, aaAo2, aaBo2, aaCo2 = aaTo, aaAo, aaBo, aaCo
        
        elif is_key_2:
            aaIntCT2, aaIntCA2, aaIntCC2 = aaIntCT, aaIntCA, aaIntCC
            aaTo2, aaAo2, aaCo2 = aaTo, aaAo, aaCo  
        
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
        
        elif is_key_3:
            aaIntCT2, aaIntCB2, aaIntCC2 = aaIntCT, aaIntCB, aaIntCC
            aaTo2, aaBo2, aaCo2 = aaTo, aaBo, aaCo
        
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
        
        elif is_key_4:
            aaIntCT2, aaIntCC2 = aaIntCT, aaIntCC
            aaTo2, aaCo2 = aaTo, aaCo
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
        
        elif is_key_5:
            aaIntCT2, aaIntCA2, aaIntCB2 = aaIntCT, aaIntCA, aaIntCB
            aaTo2, aaAo2, aaBo2 = aaTo, aaAo, aaBo
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)

        elif is_key_6:
            aaIntCT2, aaIntCA2 = aaIntCT, aaIntCA
            aaTo2, aaAo2 = aaTo, aaAo
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
        
        elif is_key_7:
            aaIntCT2, aaIntCB2 = aaIntCT, aaIntCB
            aaTo2, aaBo2 = aaTo, aaBo
            
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
            
        elif is_key_8:
            aaIntCT2 = aaIntCT
            aaTo2 = aaTo
            
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
        
        elif is_key_9:
            aaIntCA2, aaIntCB2, aaIntCC2 = aaIntCA, aaIntCB, aaIntCC
            aaAo2, aaBo2, aaCo2 = aaAo, aaBo, aaCo
            
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)

        elif is_key_10:
            aaIntCA2, aaIntCC2 = aaIntCA, aaIntCC
            aaAo2, aaCo2 = aaAo, aaCo
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
            
        elif is_key_11:
            aaIntCB2, aaIntCC2 = aaIntCB, aaIntCC
            aaBo2, aaCo2 = aaBo, aaCo
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
        
        elif is_key_12:
            aaIntCC2 = aaIntCC
            aaCo2 = aaCo
            
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)

        elif is_key_13:
            aaIntCA2, aaIntCB2 = aaIntCA, aaIntCB 
            aaAo2, aaBo2 = aaAo, aaBo
            
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
            
        elif is_key_14:
            aaIntCA2 = aaIntCA
            aaAo2 = aaAo
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
            
        elif is_key_15:
            aaIntCB2 = aaIntCB
            aaBo2 = aaBo
        
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)
            
        elif is_key_16:
            aaIntCT2, aaTo2 = process_list(aaIntCT, aaTo)
            aaIntCA2, aaAo2 = process_list(aaIntCA, aaAo)
            aaIntCB2, aaBo2 = process_list(aaIntCB, aaBo)
            aaIntCC2, aaCo2 = process_list(aaIntCC, aaCo)

        # Convert data lists to NumPy arrays and fix specific value
        aaAo2 = ['-0.000002' if x == '-2.4319000000000003e-' else x for x in aaAo2]
        data = [aaIntalpha, aaIntCS, aaIntCT2, aaIntCA2, aaIntCB2, aaIntCC2, aaSo, aaTo2, aaAo2, aaBo2, aaCo2]
        aaIal, aaICS, aaICT, aaICA, aaICB, aaICC, aaS, aaT, aaA, aaB, aaC = map(lambda x: np.array(x, dtype=float), data)
        
        # Compute `aaEst`, the estimate for Atlantic and Arctic
        aaEst = np.array([a + b*c + d*e + f*g + h*i + j*k
                          for a, b, c, d, e, f, g, h, i, j, k
                          in zip(aaIal, aaICS, aaS, aaICT, aaT, aaICA, aaA, aaICB, aaB, aaICC, aaC)])    
            
        # Store results
        aaInterpolants[key] = (aaIal, aaICS, aaICT, aaICA, aaICB, aaICC, aaEst)
            
        # Reprocessing "NaN" to 0 as needed for calculations for non-Atlantic and Arctic
        elDatao = ElseOvalues[i]
        elSo, elTo, elAo, elBo, elCo = elDatao['S'], elDatao['T'], elDatao['A'], elDatao['B'], elDatao['C']
            
        # Determine which values to use
        if is_key_1:
            elIntCT2, elIntCA2, elIntCB2, elIntCC2 = elIntCT, elIntCA, elIntCB, elIntCC
            elTo2, elAo2, elBo2, elCo2 = elTo, elAo, elBo, elCo
            
        elif is_key_2:
            elIntCT2, elIntCA2, elIntCC2 = elIntCT, elIntCA, elIntCC
            elTo2, elAo2, elCo2 = elTo, elAo, elCo
        
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
        
        elif is_key_3:
            elIntCT2, elIntCB2, elIntCC2 = elIntCT, elIntCB, elIntCC
            elTo2, elBo2, elCo2 = elTo, elBo, elCo
        
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
                          
        elif is_key_4:
            elIntCT2, elIntCC2 = elIntCT, elIntCC
            elTo2, elCo2 = elTo, elCo
            
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
        
        elif is_key_5:
            elIntCT2, elIntCA2, elIntCB2 = elIntCT, elIntCA, elIntCB
            elTo2, elAo2, elBo2 = elTo, elAo, elBo
            
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
            
        elif is_key_6:
            elIntCT2, elIntCA2 = elIntCT, elIntCA
            elTo2, elAo2 = elTo, elAo
        
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
        
        elif is_key_7:
            elIntCT2, elIntCB2 = elIntCT, elIntCB 
            elTo2, elBo2 = elTo, elBo
            
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
            
        elif is_key_8:
            elIntCT2 = elIntCT
            elTo2 = elTo
            
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
            
        elif is_key_9:
            elIntCA2, elIntCB2, elIntCC2 = elIntCA, elIntCB, elIntCC
            elAo2, elBo2, elCo2 = elAo, elBo, elCo
        
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            
        elif is_key_10:
            elIntCA2, elIntCC2 = elIntCA, elIntCC
            elAo2, elCo2 = elAo, elCo
        
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            
        elif is_key_11:
            elIntCB2, elIntCC2 = elIntCB, elIntCC
            elBo2, elCo2 = elBo, elCo
            
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            
        elif is_key_12:
            elIntCC2 = elIntCC
            elCo2 = elCo
            
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            
        elif is_key_13:
            elIntCA2, elIntCB2 = elIntCA, elIntCB
            elAo2, elBo2 = elAo, elBo
        
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
        
        elif is_key_14:
            elIntCA2 = elIntCA
            elAo2 = elAo
        
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
            
        elif is_key_15:
            elIntCB2 = elIntCB
            elBo2 = elBo
            
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)
            
        elif is_key_16:
            
            elIntCT2, elTo2 = process_list(elIntCT, elTo)
            elIntCA2, elAo2 = process_list(elIntCA, elAo)
            elIntCB2, elBo2 = process_list(elIntCB, elBo)
            elIntCC2, elCo2 = process_list(elIntCC, elCo)

        # Convert all input lists to NumPy arrays in one go
        data2 = [elIntalpha, elIntCS, elIntCT2, elIntCA2, elIntCB2, elIntCC2, elSo, elTo2, elAo2, elBo2, elCo2]
        elIal, elICS, elICT, elICA, elICB, elICC, elS, elT, elA, elB, elC = map(lambda x: np.array(x, dtype=float), data2)
            
        # compute 'elEst', the estimate for not Alantic or Atctic
        elEst = np.array([a + b*c + d*e + f*g + h*i + j*k
                          for a, b, c, d, e, f, g, h, i, j, k
                          in zip(elIal, elICS, elS, elICT, elT, elICA, elA, elICB, elB, elICC, elC)])
        # Store the results
        elInterpolants[key] = (elIal, elICS, elICT, elICA, elICB, elICC, elEst)

    Estimate, CoefficientsUsed = {}, {}
    for kcombo in AAdata.keys():
        AAdata[kcombo]["C0"] = aaInterpolants[kcombo][0]
        AAdata[kcombo]["CS"] = aaInterpolants[kcombo][1] 
        AAdata[kcombo]["CT"] = aaInterpolants[kcombo][2] 
        AAdata[kcombo]["CA"] = aaInterpolants[kcombo][3] 
        AAdata[kcombo]["CB"] = aaInterpolants[kcombo][4]
        AAdata[kcombo]["CC"] = aaInterpolants[kcombo][5]
        AAdata[kcombo]["Estimate"] = aaInterpolants[kcombo][6]
    for kcombo in Elsedata.keys():
        Elsedata[kcombo]["C0"] = elInterpolants[kcombo][0]
        Elsedata[kcombo]["CS"] = elInterpolants[kcombo][1]
        Elsedata[kcombo]["CT"] = elInterpolants[kcombo][2]
        Elsedata[kcombo]["CA"] = elInterpolants[kcombo][3]
        Elsedata[kcombo]["CB"] = elInterpolants[kcombo][4] 
        Elsedata[kcombo]["CC"] = elInterpolants[kcombo][5]
        Elsedata[kcombo]["Estimate"] = elInterpolants[kcombo][6]
            
        # Merge AA and Else data by key
        merged = {}
        for key in AAdata[kcombo].keys():
            merged[key] = np.concatenate([AAdata[kcombo][key], Elsedata[kcombo][key]])
        
        # Get sort order based on "Order"
        sort_index = np.argsort(merged["Order"])
    
        # Sort each field in the merged dictionary
        TotData = {key: val[sort_index] for key, val in merged.items()}
        
        # Store estimate values as dictionary with 1 key 
        Estimate[kcombo] =  TotData["Estimate"]

        # Store coefficients as dictionary with named keys
        CoefficientsUsed[kcombo] = {
            "Intercept": TotData["C0"],
            "Coef S": TotData["CS"],
            "Coef T": TotData["CT"],
            "Coef A": TotData["CA"],
            "Coef B": TotData["CB"],
            "Coef C": TotData["CC"]
        }
        
    return Estimate, CoefficientsUsed



def errors(OutputCoordinates={}, PredictorMeasurements={}):

    """
    Custom error messages for PyESPER that check inputs and ensure that formatting and other requirements are met
    """

    # Checking for presence of required input parameters and raising a custom error message if needed
    class CustomError(Exception):
        pass

    required_coords = ("longitude", "latitude", "depth")
    for coord_name in required_coords:
        if coord_name not in OutputCoordinates:
            raise CustomError(f"Warning: Missing {coord_name} in OutputCoordinates.")

    if "salinity" not in PredictorMeasurements: 
        raise CustomError("Warning: Missing salinity measurements. Salinity is a required input.")
            
    if "oxygen" in PredictorMeasurements and "temperature" not in PredictorMeasurements:
        raise CustomError("Warning: Missing temperature measurements. Temperature is required when oxygen is provided.")

    # Check temperature sanity and print a warning for out-of-range values
    if "temperature" in PredictorMeasurements and any(t < -5 or t > 50 for t in PredictorMeasurements["temperature"]):
        print("Warning: Temperatures below -5°C or above 50°C found. PyESPER is not designed for seawater with these properties. Ensure temperatures are in Celsius.")
                
    if any(s < 5 or s > 50 for s in PredictorMeasurements["salinity"]):
        print("Warning: Salinities less than 5 or greater than 50 have been found. ESPER is not intended for seawater with these properties.")
        
    if any(d < 0 for d in OutputCoordinates["depth"]):
        print("Warning: Depth can not be negative.")
        
    if any(l > 90 for l in OutputCoordinates["latitude"]):
        print("Warning: A latitude >90 deg (N or S) has been detected. Verify latitude is entered correctly as an input.")
    
    # Checking for commonly used missing data indicator flags. Consider adding your commonly used flags here.  
    if any(l == -9999 or l == -9 or l == -1e20 for l in OutputCoordinates["latitude"]):
           print("Warning: A common non-NaN missing data indicator (e.g., -999, -9, -1e20) was detected in the input measurements provided. Missing data should be replaced with NaNs. Otherwise, ESPER will interpret your inputs at face value and give terrible estimates.")  
 
    print("Please note that, for consistency with MATLAB ESPERv1, the now-deprecated sw package is used. This will be replaced with gsw in future updates.")



def emlr_nn(Path, DesiredVariables, Equations, OutputCoordinates={}, PredictorMeasurements={}, **kwargs):

    """
    Estimating EMLR for neural networks.
    Returns a dictionary with (DesiredVariable, Equation) as keys and Uncertainties as values.
    """

    from PyESPER.fetch_polys_NN import fetch_polys_NN
    import numpy as np
    from scipy.interpolate import griddata

    EMLR = {}

    for dv in DesiredVariables:
        DV = f"{dv}"
        NN_data = fetch_polys_NN(Path, [DV])

        data_arrays = [
            np.nan_to_num(np.array([
                NN_data[1][i][c][b][a]
                for a in range(16)
                for b in range(11)
                for c in range(8)
            ]))
            for i in range(4)
        ]

        # Create Dictionary of predetermined Uncertainties
        UGridArray = {
            'UDepth': data_arrays[0],
            'USal': data_arrays[1],
            'Eqn': data_arrays[2],
            'RMSE': data_arrays[3],
        }

        UGridPoints = (UGridArray['UDepth'], UGridArray['USal'], UGridArray['Eqn'])
        UGridValues = UGridArray['RMSE']

        for eq in Equations:
            name = dv + str(eq)
            eq_array = np.full_like(OutputCoordinates['depth'], eq, dtype=float)

            # Perform estimation for each equation
            EM = griddata(
                UGridPoints,
                UGridValues,
                (OutputCoordinates['depth'], PredictorMeasurements['salinity'], eq_array),
                method='linear'
            )

            EMLR[name] = EM

    return EMLR




def emlr_estimate(Equations, DesiredVariables, Path, OutputCoordinates={}, PredictorMeasurements={}, UDict={}, DUDict={}, Coefficients={}, **kwargs):
    
    """
    Uncertainty estimation step 1 for LIRs

    Inputs:
        Equations: List of equations
        DesiredVariables: List of variables to estimate
        Path: User-defined computer path
        OutputCoordinates: Dictionary of locations where estimates are requested
        PredictorMeasurements: Dictionary of measurements provided by user
        UDict: Dictionary of user-defined measurement uncertainties
        DUDict: Dictionary of default measurement uncertainties
        Coefficients: Dictionary of dictionaries of coefficients for each
            variable-equation scenario
        **kwargs: Please see README for full description

    Output:
        EMLR: Dictionary of uncertainty values for each desired variable-equation
            case scenario and estimate
    """

    import numpy as np
    from scipy.interpolate import griddata
    from PyESPER.fetch_data import fetch_data

    # Predefine dictionary and lists to fill
    EMLR, varnames, EqM = {}, [], []

    # Iterating over variables to fetch data this time
    for dv in DesiredVariables:
        # Fetch LIR data and process into grid arrays
        LIR_data = fetch_data([dv], Path)

        # Construct the grid array
        LIR_data = fetch_data([dv], Path)

        # Some formatting of the uncertainties from the import
        arr = np.array(LIR_data)
        arr = arr[3]
        arritem = arr.item()

        UGridArray = np.array([
            np.nan_to_num([arritem[i][c][b][a] for a in range(16) for b in range(11) for c in range(8)])
            for i in range(len(arritem))
        ]).T

        # Grid columns: UDepth, USal, Eqn, RMSE
        UDepth, USal, Eqn, RMSE = UGridArray.T
        UGridPoints = (UDepth, USal, Eqn)
        UGridValues = RMSE
    
        # Iterating over equations within variables to interpolate the uncertainties to
        # desired locations
        for eq in Equations:
            varname = dv + str(eq)
            varnames.append(varname)
            EM = []

            eq_repeated = np.full_like(OutputCoordinates['depth'], eq)
            UGridPointsOut = (
                np.array(OutputCoordinates['depth']),
                np.array(PredictorMeasurements['salinity']),  
                eq_repeated
            )
            emlr = griddata(UGridPoints, UGridValues, UGridPointsOut, method='linear')

            combo = f"{dv}{eq}"
            Coefs = {
                k: np.nan_to_num(np.array(UDict[combo][k]))
                for k in ["US", "UT", "UA", "UB", "UC"]
            }
        
            uncdfs, duncdfs = UDict[combo], DUDict[combo]
        
            # Extract keys
            keys = list(uncdfs.keys())
        
            # Function to fill arrays with floats
            def safe_fill(arr, fill_val):
                arr = np.array(arr, dtype=float)
                arr[np.isnan(arr)] = fill_val
                return arr

            # Fill=-9999 if needed
            USu2 = [safe_fill(uncdfs[k], -9999.0) for k in keys]
            UTu2 = [safe_fill(uncdfs[k], -9999.0) for k in keys]
            UAu2 = [safe_fill(uncdfs[k], -9999.0) for k in keys]
            UBu2 = [safe_fill(uncdfs[k], -9999.0) for k in keys]
            UCu2 = [safe_fill(uncdfs[k], -9999.0) for k in keys]
            
            DUSu2 = [safe_fill(duncdfs[k], -9999.0) for k in keys]
            DUTu2 = [safe_fill(duncdfs[k], -9999.0) for k in keys]
            DUAu2 = [safe_fill(duncdfs[k], -9999.0) for k in keys]
            DUBu2 = [safe_fill(duncdfs[k], -9999.0) for k in keys]
            DUCu2 = [safe_fill(duncdfs[k], -9999.0) for k in keys]

           # Compute uncertainty estimates
            EM = []

            for cucombo in range(len(Coefs["US"])):
                # Grab each coefficient
                s = Coefs["US"][cucombo]
                t = Coefs["UT"][cucombo]
                a = Coefs["UA"][cucombo]
                b = Coefs["UB"][cucombo]
                c = Coefs["UC"][cucombo]

                # Main uncertainty components
                s1 = (s * USu2[0][cucombo]) ** 2
                t1 = (t * UTu2[1][cucombo]) ** 2
                a1 = (a * UAu2[2][cucombo]) ** 2
                b1 = (b * UBu2[3][cucombo]) ** 2
                c1 = (c * UCu2[4][cucombo]) ** 2
                sum2 = s1 + t1 + a1 + b1 + c1

                # Delta uncertainties
                ds1 = (s * DUSu2[0][cucombo]) ** 2
                dt1 = (t * DUTu2[1][cucombo]) ** 2
                da1 = (a * DUAu2[2][cucombo]) ** 2
                db1 = (b * DUBu2[3][cucombo]) ** 2
                dc1 = (c * DUCu2[4][cucombo]) ** 2
                dsum2 = ds1 + dt1 + da1 + db1 + dc1

               # Final uncertainty
                uncestimate = np.sqrt(sum2 - dsum2 + emlr[cucombo] ** 2)
                EM.append(uncestimate)

            EqM.append(EM)
                    
            # Post-process and apply nan masks
            EqM2 = []
            for EM_arr in EqM:
                UncertEst = np.array(EM_arr, dtype=float)
                    
                # Convert -9999 markers to np.nan based on rules
                UncertEst[USu2[0] == -9999] = np.nan
                    
                if eq in [1, 2, 3, 4, 5, 6, 7, 8]:  
                    UncertEst[UTu2[1] == -9999] = np.nan
                if eq in [1, 2, 5, 6, 9, 10, 13, 14]:
                    UncertEst[UAu2[2] == -9999] = np.nan
                if eq in [1, 3, 5, 7, 9, 11, 13, 15]: 
                    UncertEst[UBu2[3] == -9999] = np.nan
                if eq in [1, 2, 3, 4, 9, 10, 11, 12]: 
                    UncertEst[UCu2[4] == -9999] = np.nan

                EqM2.append(UncertEst)
                
            # Final assembly into dictionary
            for i, key in enumerate(varnames):
                EMLR[key] = EqM2[i]
                    
    return EMLR




from setuptools import setup, find_packages

with open("README.md", "r") as f:
    description = f.read()

setup(
    name="PyESPER",
    version="1.0.2",
    description="Python version of ESPERv1",
    author="LMD",
    author_email="lmdias@uw.edu",
    packages=["PyESPER"],
    install_requires=[
        "numpy", 
        "seawater", 
        "scipy", 
        "matplotlib", 
        "PyCO2SYS",
        "pandas"
    ],
    entry_points={
        "console_scripts": [
            "lir = PyESPER:lir",
            "nn = PyESPER:nn",
            "mixed = PyESPER:mixed",
        ],
    },
    long_description=description,
    long_description_content_type="text/markdown",
)



def fetch_polys_NN(Path, DesiredVariables):

    """
    Loads the uncertainty polygons for NNs
    """

    from scipy.io import loadmat
    import os

    for v in DesiredVariables:
        fname = os.path.join(Path, f"Uncertainty_Polys/NN_files_{v}_Unc_Poly.mat")
        NNs = loadmat(fname)
        Polys, UncGrid = NNs["Polys"][0][0], NNs["UncGrid"][0][0]

    NN_data = [Polys, UncGrid]
    return NN_data



def lir(DesiredVariables, Path, OutputCoordinates={}, PredictorMeasurements={}, **kwargs):
    
    """
    Locally Interpolated Regressions (LIRs) for Empirical Seawater Property Estimation
    Runs all associated functions that calculate seawater properties from geographic coordinates,
        salinity, and optional other input variables using interpolation methods. Please refer
        to the README for more information.

    Inputs:
        DesiredVariables: List of desired output variables
        Path: Optional change of path location relative to user computer paths
        OutputCoordinates: Dictionary of "latitude", "longitude", and "depth" location outputs
        PredictorMeasurements: Dictionary of "salinity" and other optional predictor measurements
        **kwargs include MeasUncerts, EstDates, Equations, PerKgSwTF, VerboseTF - see README for
            full explanations

    Outputs:
        Estimates: Dictionary of estimates for each desired variable - equation combination
        Coefficients: Dictionary of dictionaries of coefficients for each combination
        Uncertainties: Dictionary of uncertainties for each combination
    """

    import time
    from PyESPER.errors import errors
    from PyESPER.defaults import defaults
    from PyESPER.lir_uncertainties import measurement_uncertainty_defaults
    from PyESPER.inputdata_organize import inputdata_organize
    from PyESPER.temperature_define import temperature_define
    from PyESPER.iterations import iterations
    from PyESPER.fetch_data import fetch_data
    from PyESPER.input_AAinds import input_AAinds
    from PyESPER.coefs_AAinds import coefs_AAinds
    from PyESPER.interpolate import interpolate
    from PyESPER.organize_data import organize_data
    from PyESPER.emlr_estimate import emlr_estimate
    from PyESPER.adjust_pH_DIC import adjust_pH_DIC
    from PyESPER.pH_adjustment import pH_adjustment
    from PyESPER.pH_adjcalc import pH_adjcalc
    from PyESPER.final_formatting import final_formatting

    # Starting the timer
    tic = time.perf_counter() 
    
    # Providing custom error messages for erroneous input
    errors(OutputCoordinates, PredictorMeasurements)

    # Setting defaults for various input parameters, including defining kwargs and
    # ensuring that coordinates use the correct format
    Equations, n, VerboseTF, EstDates, C, PerKgSwTF, MeasUncerts = defaults(
        DesiredVariables, 
        PredictorMeasurements,
        OutputCoordinates, 
        **kwargs
    )
    
    # Processing the input values (Uncertainties_pre) and calculating default 
    # measurement uncertainties 
    Uncertainties_pre, DUncertainties_pre  = measurement_uncertainty_defaults(
        n, 
        PredictorMeasurements, 
        MeasUncerts
    )
    
    # Creating an updated dictionary of all input data
    InputAll  = inputdata_organize(
        EstDates, 
        C, 
        PredictorMeasurements, 
        Uncertainties_pre
    )

    # Defining temperature as needed
    PredictorMeasurements, InputAll = temperature_define(
        DesiredVariables,
        PredictorMeasurements,
        InputAll,
        **kwargs
    )

    # Performing iterations for equation-desired variable combinations;
    # pre-defines the correct input data for LIRs
    code, unc_combo_dict, dunc_combo_dict = iterations(
        DesiredVariables, 
        Equations, 
        PerKgSwTF,
        C,
        PredictorMeasurements, 
        InputAll,
        Uncertainties_pre,
        DUncertainties_pre
    )

    # Loading the pre-trained algorithm data
    LIR_data = fetch_data(
        DesiredVariables, 
        Path
    )

    # Separating user-defined coordinates into Atlantic and Arctic (AAdata)
    # or other regions (Elsedata)
    AAdata, Elsedata = input_AAinds(
        C, 
        code
    )

    # Separating ESPER pre-defined coefficients into Atlantic and Arctic or other regions
    Gdf, CsDesired = coefs_AAinds(
        Equations, 
        LIR_data
    )

    # Interpolate
    aaLCs, aaInterpolants_pre, elLCs, elInterpolants_pre = interpolate(
        Gdf, 
        AAdata, 
        Elsedata
    )
 
    # Organize data and compute estimates
    Estimate, CoefficientsUsed = organize_data(
        aaLCs, 
        elLCs, 
        aaInterpolants_pre, 
        elInterpolants_pre, 
        Gdf, 
        AAdata,
        Elsedata
    )

    # Calculate initial uncertainties for lirs
    Uncertainties = emlr_estimate(
        Equations, 
        DesiredVariables, 
        Path,
        OutputCoordinates, 
        PredictorMeasurements, 
        unc_combo_dict, 
        dunc_combo_dict,
        Coefficients=CoefficientsUsed)
   
    # First of three steps to adjust pH and DIC for
    # anthropogenic carbon, as needed
    Cant_adjusted, Cant, Cant2002 = adjust_pH_DIC(
        DesiredVariables,
        VerboseTF,
        EstDates,
        Estimate,
        PredictorMeasurements,
        OutputCoordinates,
        **kwargs
    )

    # Second of three steps for Cant adjustment, for pH only
    Cant_adjusted = pH_adjustment(
        Path,
        DesiredVariables, 
        EstDates, 
        Cant,
        Cant2002, 
        PerKgSwTF,
        Cant_adjusted, 
        Estimate,
        PredictorMeasurements,
        OutputCoordinates,
        C,
        Uncertainties_pre,
        DUncertainties_pre,
        **kwargs
    )

    # Last of three adjustments for anthropogenic carbon
    Cant_adjusted, combos2, values2 = pH_adjcalc(
        DesiredVariables,
        VerboseTF,
        Estimate,
        Cant_adjusted,
        **kwargs
    )

    # Finalizing formatting of estimate output
    Estimates = final_formatting(
        DesiredVariables, 
        Cant_adjusted, 
        Estimate
    )
   
     # Stopping the timer
    toc = time.perf_counter()
    print(f"PyESPER_LIR took {toc - tic:0.4f} seconds, or {(toc-tic)/60:0.4f} minutes to run")    

    return Estimates, CoefficientsUsed, Uncertainties
